"""
The ``main`` function here is wired to the command line tool by name
asynctelnet-server.  If this server's PID receives the SIGTERM signal, it
attempts to shutdown gracefully.

The :class:`TelnetServer` class negotiates a character-at-a-time (WILL-SGA,
WILL-ECHO) session with support for negotiation about window size, environment
variables, terminal type name, and to automatically close connections clients
after an idle period.
"""
# std imports
import collections
import argparse
import asyncio
import logging
import signal
import anyio
from weakref import proxy
from contextlib import asynccontextmanager
from functools import partial

# local
from . import server_base
from . import accessories

__all__ = ('TelnetServer', 'server_loop', 'run_server', 'parse_server_args')

CONFIG = collections.namedtuple('CONFIG', [
    'host', 'port', 'loglevel', 'logfile', 'logfmt', 'shell', 'encoding',
    'force_binary', 'timeout'])(
        host='localhost', port=6023, loglevel='info',
        logfile=None, logfmt=accessories._DEFAULT_LOGFMT ,
        shell=accessories.function_lookup('asynctelnet.telnet_server_shell'),
        encoding='utf8', force_binary=False, timeout=300)


class TelnetServer(server_base.BaseServer):
    """Telnet Server protocol performing common negotiation."""
    #: Maximum number of cycles to seek for all terminal types.  We are seeking
    #: the repeat or cycle of a terminal table, choosing the first -- but when
    #: negotiated by MUD clients, we chose the must Unix TERM appropriate,
    TTYPE_LOOPMAX = 8

    # Derived methods from base class

    def __init__(self, stream, term='unknown', cols=80, rows=25, timeout=300,
                 *args, **kwargs):
        super().__init__(stream, *args, **kwargs)
        self._ttype_count = 1
        self._timer = None
        self._extra = {
            'term': term,
            'charset': kwargs.get('encoding', ''),
            'cols': cols,
            'rows': rows,
            'timeout': timeout
        }

    @property
    def extra_attributes(self):
        res = super().extra_attributes
        for k in self._extra.keys(): 
            def fn(k):
                return lambda: self._extra[k] 
            res[k] = fn(k)
        return res

    async def setup(self):
        from .telopt import NAWS, NEW_ENVIRON, TSPEED, TTYPE, XDISPLOC, CHARSET
        await super().setup()

        # begin timeout timer
#       async with self.with_timeout():

#           # Wire extended rfc callbacks for responses to
#           # requests of terminal attributes, environment values, etc.
#           for tel_opt, callback_fn in [
#               (NAWS, self.on_naws),
#               (NEW_ENVIRON, self.on_environ),
#               (TSPEED, self.on_tspeed),
#               (TTYPE, self.on_ttype),
#               (XDISPLOC, self.on_xdisploc),
#               (CHARSET, self.on_charset),
#           ]:
#               self.writer.set_ext_callback(tel_opt, callback_fn)

#           # Wire up a callbacks that return definitions for requests.
#           for tel_opt, callback_fn in [
#               (NEW_ENVIRON, self.on_request_environ),
#               (CHARSET, self.on_request_charset),
#           ]:
#               self.writer.set_ext_send_callback(tel_opt, callback_fn)


        from .telopt import (DO, WILL, SGA, ECHO, BINARY, TTYPE, 
                             NEW_ENVIRON, NAWS, CHARSET)
        # No terminal? don't try.
        if await self.remote_option(TTYPE, True):
            async with anyio.create_task_group() as tg:
                await tg.spawn(self.local_option,SGA,True)
                await tg.spawn(self.local_option,ECHO,True)
                await tg.spawn(self.local_option,BINARY,True)
                await tg.spawn(self.remote_option,NEW_ENVIRON,True)
                await tg.spawn(self.remote_option,NAWS,True)
                await tg.spawn(self.remote_option,BINARY,True)
                await tg.spawn(self.request_charset("UTF-8"))

        # TODO request environment

        # prefer 'LANG' environment variable forwarded by client, if any.
        # for modern systems, this is the preferred method of encoding
        # negotiation.
        #_lang = self.get_extra_info('LANG', '')
        #if _lang:
        #    return accessories.encoding_from_lang(_lang)

        # otherwise, the less CHARSET negotiation may be found in many
        # East-Asia BBS and Western MUD systems.
        #return self.get_extra_info('charset') or self.default_encoding

    @asynccontextmanager
    async def with_timeout(self, duration=-1):
        """
        Restart or unset timeout for client.

        :param int duration: When specified as a positive integer,
            schedules Future for callback of :meth:`on_timeout`.  When ``-1``,
            the value of ``self.get_extra_info('timeout')`` is used.  When
            non-True, it is canceled.
        """
        if duration == -1:
            duration = self.get_extra_info('timeout')
        if duration > 0:
            async with anyio.move_on_after(duration) as sc:
                yield sc
        else:
            yield None

    def on_naws(self, rows, cols):
        """
        Callback receives NAWS response, :rfc:`1073`.

        :param int rows: screen size, by number of cells in height.
        :param int cols: screen size, by number of cells in width.
        """
        self._extra.update({'rows': rows, 'cols': cols})

    def on_request_environ(self):
        """
        Definition for NEW_ENVIRON request of client, :rfc:`1572`.

        This method is a callback from :meth:`~.TelnetWriter.request_environ`,
        first entered on receipt of (WILL, NEW_ENVIRON) by server.  The return
        value *defines the request made to the client* for environment values.

        :rtype list: a list of unicode character strings of US-ASCII
            characters, indicating the environment keys the server requests
            of the client.  If this list contains the special byte constants,
            ``USERVAR`` or ``VAR``, the client is allowed to volunteer any
            other additional user or system values.

            Any empty return value indicates that no request should be made.

        The default return value is::

            ['LANG', 'TERM', 'COLUMNS', 'LINES', 'DISPLAY', 'COLORTERM',
             VAR, USERVAR, 'COLORTERM']
        """
        from .telopt import VAR, USERVAR
        return ['LANG', 'TERM', 'COLUMNS', 'LINES', 'DISPLAY', 'COLORTERM',
                VAR, USERVAR]

    def on_environ(self, mapping):
        """Callback receives NEW_ENVIRON response, :rfc:`1572`."""
        # A well-formed client responds with empty values for variables to
        # mean "no value".  They might have it, they just may not wish to
        # divulge that information.  We pop these keys as a side effect in
        # the result statement of the following list comprehension.
        no_value = [mapping.pop(key) or key
                    for key, val in list(mapping.items())
                    if not val]

        # because we are working with "untrusted input", we make one fair
        # distinction: all keys received by NEW_ENVIRON are in uppercase.
        # this ensures a client may not override trusted values such as
        # 'peer'.
        u_mapping = {key.upper(): val for key, val in list(mapping.items())}

        self.log.debug('on_environ received: {0!r}'.format(u_mapping))

        self._extra.update(u_mapping)

    def on_request_charset(self):
        """
        Definition for CHARSET request by client, :rfc:`2066`.

        This method is a callback from :meth:`~.TelnetWriter.request_charset`,
        first entered on receipt of (WILL, CHARSET) by server.  The return
        value *defines the request made to the client* for encodings.

        :rtype list: a list of unicode character strings of US-ASCII
            characters, indicating the encodings offered by the server in
            its preferred order.

            Any empty return value indicates that no encodings are offered.

        The default return value begins::

            ['UTF-8', 'UTF-16', 'LATIN1', 'US-ASCII', 'BIG5', 'GBK', ...]
        """
        return ['UTF-8', 'UTF-16', 'LATIN1', 'US-ASCII', 'BIG5',
                'GBK', 'SHIFTJIS', 'GB18030', 'KOI8-R', 'KOI8-U',
                ] + [
                    # "Part 12 was slated for Latin/Devanagari,
                    # but abandoned in 1997"
                    'ISO8859-{}'.format(iso) for iso in range(1, 16)
                    if iso != 12
                ] + ['CP{}'.format(cp) for cp in (
                    154, 437, 500, 737, 775, 850, 852, 855, 856, 857,
                    860, 861, 862, 863, 864, 865, 866, 869, 874, 875,
                    932, 949, 950, 1006, 1026, 1140, 1250, 1251, 1252,
                    1253, 1254, 1255, 1257, 1257, 1258, 1361,
                )]

    def on_charset(self, charset):
        """Callback for CHARSET response, :rfc:`2066`."""
        self._extra['charset'] = charset

    def on_tspeed(self, rx, tx):
        """Callback for TSPEED response, :rfc:`1079`."""
        self._extra['tspeed'] = '{0},{1}'.format(rx, tx)

    def on_ttype(self, ttype):
        """Callback for TTYPE response, :rfc:`930`."""
        # TTYPE may be requested multiple times, we honor this system and
        # attempt to cause the client to cycle, as their first response may
        # not be their most significant. All responses held as 'ttype{n}',
        # where {n} is their serial response order number.
        #
        # The most recently received terminal type by the server is
        # assumed TERM by this implementation, even when unsolicited.
        key = 'ttype{}'.format(self._ttype_count)
        self._extra[key] = ttype
        if ttype:
            self._extra['TERM'] = ttype

        _lastval = self.get_extra_info('ttype{0}'.format(
            self._ttype_count - 1))

        if key != 'ttype1' and ttype == self.get_extra_info('ttype1', None):
            # cycle has looped, stop
            self.log.debug('ttype cycle stop at {0}: {1}, looped.'
                           .format(key, ttype))

        elif (not ttype or self._ttype_count > self.TTYPE_LOOPMAX):
            # empty reply string or too many responses!
            self.log.warning('ttype cycle stop at {0}: {1}.'.format(key, ttype))

        elif (self._ttype_count == 3 and ttype.upper().startswith('MTTS ')):
            val = self.get_extra_info('ttype2')
            self.log.debug(
                'ttype cycle stop at {0}: {1}, using {2} from ttype2.'
                .format(key, ttype, val))
            self._extra['TERM'] = val

        elif (ttype == _lastval):
            self.log.debug('ttype cycle stop at {0}: {1}, repeated.'
                           .format(key, ttype))

        else:
            self.log.debug('ttype cycle cont at {0}: {1}.'
                           .format(key, ttype))
            self._ttype_count += 1
            self.writer.request_ttype()

    def on_xdisploc(self, xdisploc):
        """Callback for XDISPLOC response, :rfc:`1096`."""
        self._extra['xdisploc'] = xdisploc


async def server_loop(host=None, port=23, evt=None, protocol_factory=TelnetServer, shell=None, log=None, **kwds):
    """
    Run a TCP Telnet server.

    :param str host: The host parameter can be a string, in that case the TCP
        server is bound to host and port. The host parameter can also be a
        sequence of strings, and in that case the TCP server is bound to all
        hosts of the sequence.
    :param int port: listen port for TCP Server.
    :param server_base.BaseServer protocol_factory: An alternate protocol
        factory for the server, when unspecified, :class:`TelnetServer` is
        used.

    :param Callable shell: A coroutine that is called after
        negotiation completes, receiving the Telnet stream as an argument.
    :param logging.Logger log: target logger, if None is given, one is created
        using the namespace ``'asynctelnet.server'``.
    :param str encoding: The default assumed encoding, or ``False`` to disable
        unicode support.  Encoding may be negotiation to another value by
        the client through NEW_ENVIRON :rfc:`1572` by sending environment value
        of ``LANG``, or by any legal value for CHARSET :rfc:`2066` negotiation.

        The server's stream accepts and returns Unicode, unless this value
        is explicitly set to ``None``.  In that case, the attached stream
        interface is bytes-only.
    :param str encoding_errors: Same meaning as :meth:`codecs.Codec.encode`.
        Default value is ``strict``.
    :param bool force_binary: When ``True``, the encoding specified is
        used for both directions even when BINARY mode, :rfc:`856`, is not
        negotiated for the direction specified.  This parameter has no effect
        when ``encoding=None``.
    :param str term: Value returned for ``writer.get_extra_info('term')``
        until negotiated by TTYPE :rfc:`930`, or NAWS :rfc:`1572`.  Default value
        is ``'unknown'``.
    :param int cols: Value returned for ``writer.get_extra_info('cols')``
        until negotiated by NAWS :rfc:`1572`. Default value is 80 columns.
    :param int rows: Value returned for ``writer.get_extra_info('rows')``
        until negotiated by NAWS :rfc:`1572`. Default value is 25 rows.
    :param float connect_maxwait: If the remote end is not compliant, or
        otherwise confused by our demands, the shell continues anyway after the
        greater of this value has elapsed.  A client that is not answering
        option negotiation will delay the start of the shell by this amount.

    This method does not return until cancelled.
    """
    protocol_factory = protocol_factory or TelnetServer
    l = await anyio.create_tcp_listener(local_host=host, local_port=port)
    if shell is None:
        async def shell(_s):
            while True:
                await anyio.sleep(99999)
    async def serve(s):
        async with protocol_factory(s, log=log, **kwds) as stream:
            await shell(stream)

    if log is not None:
        log.info('Server ready on {0}:{1}'.format(host, port))
    if evt is not None:
        await evt.set()
    await l.serve(serve)


async def _sigterm_handler(server, log):
    log.info('SIGTERM received, closing server.')

    # This signals the completion of the server.wait_closed() Future,
    # allowing the main() function to complete.
    server.close()


def parse_server_args():
    parser = argparse.ArgumentParser(
        description="Telnet protocol server",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('host', nargs='?', default=CONFIG.host,
                        help='bind address')
    parser.add_argument('port', nargs='?', type=int, default=CONFIG.port,
                        help='bind port')
    parser.add_argument('--loglevel', default=CONFIG.loglevel,
                        help='level name')
    parser.add_argument('--logfile', default=CONFIG.logfile,
                        help='filepath')
    parser.add_argument('--logfmt', default=CONFIG.logfmt,
                        help='log format')
    parser.add_argument('--shell', default=CONFIG.shell,
                        type=accessories.function_lookup,
                        help='module.function_name')
    parser.add_argument('--encoding', default=CONFIG.encoding,
                        help='encoding name')
    parser.add_argument('--force-binary', action='store_true',
                        default=CONFIG.force_binary,
                        help='force binary transmission')
    parser.add_argument('--timeout', default=CONFIG.timeout,
                        help='idle disconnect (0 disables)')
#   parser.add_argument('--connect-maxwait', type=float,
#                       default=CONFIG.connect_maxwait,
#                       help='timeout for pending negotiation')
    return vars(parser.parse_args())


def run_server(host=CONFIG.host, port=CONFIG.port, loglevel=CONFIG.loglevel,
               logfile=CONFIG.logfile, logfmt=CONFIG.logfmt,
               shell=CONFIG.shell, encoding=CONFIG.encoding,
               force_binary=CONFIG.force_binary, timeout=CONFIG.timeout,
#              connect_maxwait=CONFIG.connect_maxwait
                ):
    """
    Program entry point for server daemon.

    This function configures a logger and creates a telnet server for the
    given keyword arguments, serving forever, completing only upon receipt of
    SIGTERM.
    """
    log = accessories.make_logger(
        name=__name__,
        loglevel=loglevel,
        logfile=logfile,
        logfmt=logfmt)

    anyio.run(partial(
        server_loop, host, port, shell=shell, encoding=encoding,
                      force_binary=force_binary, timeout=timeout, log=log))


    # await completion of server stop
    try:
        loop.run_until_complete(server.wait_closed())
    finally:
        # remove signal handler on stop
        loop.remove_signal_handler(signal.SIGTERM)

    log.info('Server stop.')


def main():
    """Command-line 'asynctelnet-server' entry point, via setuptools."""
    return run_server(**parse_server_args())


if __name__ == '__main__':
    exit(main())



import anyio
from anyio.streams.text import TextStream

CR, LF, NUL = '\r\n\x00'
from . import slc
from . import telopt
from . import accessories

__all__ = ('telnet_server_shell',)


async def telnet_server_shell(stream):
    """
    A default telnet shell, appropriate for use with asynctelnet.create_server.

    This shell provides a very simple REPL, allowing introspection and state
    toggling of the connected client session.

    """
    stream = RLTextStream(stream)
    await stream.send("Ready." + CR + LF)

    command = None
    while True:
        if command:
            await writer.send(CR + LF)
        await writer.send('tel:sh> ')
        command = await stream.readline()
        if not command:
            return
        await stream.send(CR + LF)
        if command == 'quit':
            await stream.send('Goodbye.' + CR + LF)
            break
        elif command == 'help':
            await stream.send('quit, stream, slc, toggle [option|all], '
                         '… and that\'s it.')
        elif command == 'stream':
            await stream.send(repr(stream))
        elif command == 'version':
            await stream.send(accessories.get_version())
        elif command == 'slc':
            await stream.send(get_slcdata(stream))
        elif command.startswith('toggle'):
            option = command[len('toggle '):] or None
            await stream.send(await do_toggle(stream, option))
        elif command:
            await stream.send(f'no such command: {command!r}.')
    await stream.aclose()


def get_slcdata(writer):
    """Display Special Line Editing (SLC) characters."""
    _slcs = sorted([
        '{:>15}: {}'.format(slc.name_slc_command(slc_func), slc_def)
        for (slc_func, slc_def) in sorted(writer.slctab.items())
        if not (slc_def.nosupport or slc_def.val == slc.theNULL)])
    _unset = sorted([
        slc.name_slc_command(slc_func)
        for (slc_func, slc_def) in sorted(writer.slctab.items())
        if slc_def.val == slc.theNULL])
    _nosupport = sorted([
        slc.name_slc_command(slc_func)
        for (slc_func, slc_def) in sorted(
            writer.slctab.items())
        if slc_def.nosupport])

    return ('Special Line Characters:\r\n' +
            '\r\n'.join(_slcs) +
            '\r\nUnset by client: ' +
            ', '.join(_unset) +
            '\r\nNot supported by server: ' +
            ', '.join(_nosupport))


async def do_toggle(writer, option):
    """Display or toggle telnet session parameters."""
    tbl_opt = {
        'echo': writer.local_option.enabled(telopt.ECHO),
        'goahead': not writer.local_option.enabled(telopt.SGA),
        'outbinary': writer.outbinary,
        'inbinary': writer.inbinary,
        'binary': writer.outbinary and writer.inbinary,
        'xon-any': writer.xon_any,
        'lflow': writer.lflow,
    }

    if not option:
        return ('\r\n'.join('{0} {1}'.format(
            opt, 'ON' if enabled else 'off')
            for opt, enabled in sorted(tbl_opt.items())))

    msgs = []
    if option in ('echo', 'all'):
        cmd = (telopt.WONT if tbl_opt['echo'] else telopt.WILL)
        await writer.iac(cmd, telopt.ECHO)
        msgs.append('{} echo.'.format(
            telopt.name_command(cmd).lower()))

    if option in ('goahead', 'all'):
        cmd = (telopt.WILL if tbl_opt['goahead'] else telopt.WONT)
        await writer.iac(cmd, telopt.SGA)
        msgs.append('{} suppress go-ahead.'.format(
            telopt.name_command(cmd).lower()))

    if option in ('outbinary', 'binary', 'all'):
        cmd = (telopt.WONT if tbl_opt['outbinary'] else telopt.WILL)
        await writer.iac(cmd, telopt.BINARY)
        msgs.append('{} outbinary.'.format(
            telopt.name_command(cmd).lower()))

    if option in ('inbinary', 'binary', 'all'):
        cmd = (telopt.DONT if tbl_opt['inbinary'] else telopt.DO)
        await writer.iac(cmd, telopt.BINARY)
        msgs.append('{} inbinary.'.format(
            telopt.name_command(cmd).lower()))

    if option in ('xon-any', 'all'):
        writer.xon_any = not tbl_opt['xon-any']
        await writer.send_lineflow_mode()
        msgs.append('xon-any {}abled.'.format(
            'en' if writer.xon_any else 'dis'))

    if option in ('lflow', 'all'):
        writer.lflow = not tbl_opt['lflow']
        await writer.send_lineflow_mode()
        msgs.append('lineflow {}abled.'.format(
            'en' if writer.lflow else 'dis'))

    if option not in tbl_opt and option != 'all':
        msgs.append('toggle: not an option.')


    return '\r\n'.join(msgs)

class ReadLine:
    """
    A very primitive readline wrapper.

    Usage::
        rl = ReadLine(stream)
        while True:
            line = await rl()
    """
    def __init__(self, stream):
        self._stream = stream
        self._buf = ''
        self._echo = self.extra_attributes[EchoAttr]
        self._last_inp = ''

    async def __call__(self):
        buf = self._buf
        command = ""
        while True:
            if buf == "":
                buf = await self._stream.receive()
            for i,inp in enumerate(buf):
                if inp in (LF, NUL) and self._last_inp == CR:
                    self._buf = buf[i+1:]
                    return command

                elif inp in (CR, LF):
                    # first CR or LF yields command
                    self._buf = buf[i+1:]
                    return command

                elif inp in ('\b', '\x7f'):
                    # backspace over input
                    if command:
                        command = command[:-1]
                        await self.send('\b \b')

                else:
                    # buffer and echo input
                    command += inp
                    await self._stream.echo(inp)
                    self._last_inp = inp




"""
КОСМОС/300 Lunar server: a demonstration Telnet Server shell.

With this shell, multiple clients may instruct the lander to:

    - collect rock samples
    - launch sample return capsule
    - relay a message (talker).

All input and output is forced uppercase, but unicode is otherwise supported.

A simple character-at-a-time repl is provided, supporting backspace.
"""
# std imports
import collections
import contextlib
import logging
import asyncio

class Client(collections.namedtuple(
    'Client', ['reader', 'writer', 'notify_queue'])):

    def __str__(self):
        return '#{1}'.format(*self.writer.get_extra_info('peername'))

class Lander(object):
    """
    КОСМОС/300 Lunar module.
    """
    collecting = False
    capsule_amount = 0
    capsule_launched = False
    capsule_launching = False

    def __init__(self):
        self.log = logging.getLogger('lunar.lander')
        self.clients = []

    def __str__(self):
        collector = 'RUNNING' if self.collecting else 'READY'
        capsule = ('LAUNCH IN-PROGRESS' if self.capsule_launching else
                   'LAUNCHED' if self.capsule_launched else
                   '{}/4'.format(self.capsule_amount))
        clients = ', '.join(map(str, self.clients))
        return ('COLLECTOR {}\r\nCAPSULE {}\r\nUPLINKS: {}'
                .format(collector, capsule, clients))

    @contextlib.contextmanager
    def register_link(self, reader, writer):
        client = Client(reader, writer, notify_queue=asyncio.Queue())
        self.clients.append(client)
        try:
            self.notify_event('LINK ESTABLISHED TO {}'.format(client))
            yield client

        finally:
            self.clients.remove(client)
            self.notify_event('LOST CONNECTION TO {}'.format(client))

    def notify_event(self, event_msg):
        self.log.info(event_msg)
        for client in self.clients:
            client.notify_queue.put_nowait(event_msg)

    def repl_readline(self, client):
        """
        Lander REPL, provides no process, local echo.
        """
        from asynctelnet import WONT, ECHO, SGA
        client.writer.iac(WONT, ECHO)
        client.writer.iac(WONT, SGA)
        readline = asyncio.ensure_future(client.reader.readline())
        recv_msg = asyncio.ensure_future(client.notify_queue.get())
        client.writer.write('КОСМОС/300: READY\r\n')
        wait_for = set([readline, recv_msg])
        try:
            while True:
                client.writer.write('? ')

                # await (1) client input or (2) system notification
                done, pending = await asyncio.wait(
                    wait_for, return_when=asyncio.FIRST_COMPLETED)

                task = done.pop()
                wait_for.remove(task)
                if task == readline:
                    # (1) client input
                    cmd = (task.result()
                           .rstrip()
                           .upper())

                    client.writer.echo(cmd)
                    self.process_command(client, cmd)

                    # await next,
                    readline = asyncio.ensure_future(client.reader.readline())
                    wait_for.add(readline)

                else:
                    # (2) system notification
                    msg = task.result()

                    # await next,
                    recv_msg = asyncio.ensure_future(client.notify_queue.get())
                    wait_for.add(recv_msg)

                    # show and display prompt,
                    client.writer.write('\r\x1b[K{}\r\n'.format(msg))

        finally:
            for task in wait_for:
                task.cancel()


    async def repl_catime(self, client):
        """
        Lander REPL providing character-at-a-time processing.
        """
        read_one = asyncio.ensure_future(client.reader.read(1))
        recv_msg = asyncio.ensure_future(client.notify_queue.get())
        wait_for = set([read_one, recv_msg])

        client.writer.write('КОСМОС/300: READY\r\n')

        while True:
            cmd = ''

            # prompt
            client.writer.write('? ')
            while True:

                # await (1) client input (2) system notification
                done, pending = await asyncio.wait(
                    wait_for, return_when=asyncio.FIRST_COMPLETED)

                task = done.pop()
                wait_for.remove(task)
                if task == read_one:
                    # (1) client input
                    char = task.result().upper()

                    # await next,
                    read_one = asyncio.ensure_future(client.reader.read(1))
                    wait_for.add(read_one)

                    if char == '':
                        # disconnect, exit
                        return

                    elif char in '\r\n':
                        # carriage return, process command.
                        break

                    elif char in '\b\x7f':
                        # backspace
                        cmd = cmd[:-1]
                        client.writer.echo('\b')

                    else:
                        # echo input
                        cmd += char
                        client.writer.echo(char)

                else:
                    # (2) system notification
                    msg = task.result()

                    # await next,
                    recv_msg = asyncio.ensure_future(client.notify_queue.get())
                    wait_for.add(recv_msg)

                    # show and display prompt,
                    client.writer.write('\r\x1b[K{}\r\n'.format(msg))
                    client.writer.write('? {}'.format(cmd))

            # reached when user pressed return by inner 'break' statement.
            self.process_command(client, cmd)

    def process_command(self, client, cmd):
        result = '\r\n'
        if cmd == 'HELP':
            result += ('COLLECT  COLLECT ROCK SAMPLE\r\n'
                       ' STATUS  DEVICE STATUS\r\n'
                       ' LAUNCH  LAUNCH RETURN CAPSULE\r\n'
                       '  RELAY  MESSAGE TRANSMISSION RELAY')
        elif cmd == 'STATUS':
            result += str(self)
        elif cmd == 'COLLECT':
            result += self.collect_sample(client)
        elif cmd == 'LAUNCH':
            result += self.launch_capsule(client)
        elif (cmd == 'RELAY' or
              cmd.startswith('RELAY ') or
              cmd.startswith('R ')):
            cmd, *args = cmd.split(None, 1)
            if args:
                self.notify_event('RELAY FROM {}: {}'.format(client, args[0]))
            result = ''
        elif cmd:
            result += "NOT A COMMAND, {!r}".format(cmd)
        client.writer.write(result + '\r\n')

    def launch_capsule(self, client):
        if self.capsule_launched:
            return 'ERROR: NO CAPSULE'
        elif self.capsule_launching:
            return 'ERROR: LAUNCH SEQUENCE IN-PROGRESS'
        elif self.collecting:
            return 'ERROR: COLLECTOR ACTIVE'
        self.capsule_launching = True
        self.notify_event('CAPSULE LAUNCH SEQUENCE INITIATED!')
        asyncio.get_event_loop().call_later(10, self.after_launch)
        for count in range(1, 10):
            asyncio.get_event_loop().call_later(
                count, self.notify_event, '{} ...'.format(10 - count))
        return 'OK'

    def collect_sample(self, client):
        if self.collecting:
            return 'ERROR: COLLECTION ALREADY IN PROGRESS'
        elif self.capsule_launched:
            return 'ERROR: COLLECTOR CAPSULE NOT CONNECTED'
        elif self.capsule_launching:
            return 'ERROR: LAUNCH SEQUENCE IN-PROGRESS.'
        elif self.capsule_amount >= 4:
            return 'ERROR: COLLECTOR CAPSULE FULL'
        self.collecting = True
        self.notify_event('SAMPLE COLLECTION HAS BEGUN')
        self._loop.call_later(7, self.collected_sample)
        return 'OK'

    def collected_sample(self):
        self.notify_event('SAMPLE COLLECTED')
        self.capsule_amount += 1
        self.collecting = False

    def after_launch(self):
        self.capsule_launching = False
        self.capsule_launched = True
        self.notify_event('CAPSULE LAUNCHED SUCCESSFULLY')


# each client shares, even communicates through lunar 'lander' instance.
lander = Lander()

def shell(reader, writer):
    global lander
    with lander.register_link(reader, writer) as client:
        await lander.repl_readline(client)



"""Test TSPEED, rfc-1079_."""
# std imports
import asyncio

# local imports
import asynctelnet
from asynctelnet.tests.accessories import (
    unused_tcp_port,
    bind_host
)

# 3rd party
import pytest


@pytest.mark.anyio
async def test_telnet_server_on_tspeed(bind_host, unused_tcp_port):
    """Test Server's callback method on_tspeed()."""
    # given
    from asynctelnet.telopt import IAC, WILL, SB, SE, IS, TSPEED
    _waiter = asyncio.Future()

    class ServerTestTspeed(asynctelnet.TelnetServer):
        def on_tspeed(self, rx, tx):
            super().on_tspeed(rx, tx)
            _waiter.set_result(self)

    await asynctelnet.create_server(
        protocol_factory=ServerTestTspeed,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    writer.write(IAC + WILL + TSPEED)
    writer.write(IAC + SB + TSPEED + IS + b'123,456' + IAC + SE)

    # verify,
    srv_instance = await asyncio.wait_for(_waiter, 0.5)
    assert srv_instance.get_extra_info('tspeed') == '123,456'


@pytest.mark.anyio
async def test_telnet_client_send_tspeed(bind_host, unused_tcp_port):
    """Test Client's callback method send_tspeed()."""
    # given
    _waiter = asyncio.Future()
    given_rx, given_tx = 1337, 1919

    class ServerTestTspeed(asynctelnet.TelnetServer):
        def on_tspeed(self, rx, tx):
            super().on_tspeed(rx, tx)
            _waiter.set_result((rx, tx))

        def begin_advanced_negotiation(self):
            from asynctelnet.telopt import DO, TSPEED
            super().begin_advanced_negotiation()
            self.writer.iac(DO, TSPEED)

    await asynctelnet.create_server(
        protocol_factory=ServerTestTspeed,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        tspeed=(given_rx, given_tx), connect_minwait=0.05)

    recv_rx, recv_tx = await asyncio.wait_for(_waiter, 0.5)
    assert recv_rx == given_rx
    assert recv_tx == given_tx



"""Negotiate About Window Size, *NAWS*. rfc-1073_."""
# std imports
import asyncio
import pexpect
import struct

# local imports
import asynctelnet
from asynctelnet.tests.accessories import (
    unused_tcp_port,
    bind_host
)

# 3rd party
import pytest


@pytest.mark.anyio
async def test_telnet_server_on_naws(
        bind_host, unused_tcp_port):
    """Test Server's Negotiate about window size (NAWS)."""
    # given
    from asynctelnet.telopt import IAC, WILL, SB, SE, NAWS
    _waiter = asyncio.Future()
    given_cols, given_rows = 40, 20

    class ServerTestNaws(asynctelnet.TelnetServer):
        def on_naws(self, width, height):
            super().on_naws(width, height)
            _waiter.set_result(self)

    await asynctelnet.create_server(
        protocol_factory=ServerTestNaws,
        host=bind_host, port=unused_tcp_port,
        connect_maxwait=0.05)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    writer.write(IAC + WILL + NAWS)
    writer.write(IAC + SB + NAWS +
                 struct.pack('!HH', given_cols, given_rows) +
                 IAC + SE)

    srv_instance = await asyncio.wait_for(_waiter, 0.5)
    assert srv_instance.get_extra_info('cols') == given_cols
    assert srv_instance.get_extra_info('rows') == given_rows


@pytest.mark.anyio
async def test_telnet_client_send_naws(bind_host, unused_tcp_port):
    """Test Client's NAWS of callback method send_naws()."""
    # given a server
    _waiter = asyncio.Future()
    given_cols, given_rows = 40, 20

    class ServerTestNaws(asynctelnet.TelnetServer):
        def on_naws(self, width, height):
            super().on_naws(width, height)
            _waiter.set_result((height, width))

    await asynctelnet.create_server(
        protocol_factory=ServerTestNaws,
        host=bind_host, port=unused_tcp_port,
        connect_maxwait=0.05)

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        cols=given_cols, rows=given_rows, connect_minwait=0.05)

    recv_cols, recv_rows = await asyncio.wait_for(_waiter, 0.5)
    assert recv_cols == given_cols
    assert recv_rows == given_rows


@pytest.mark.anyio
async def test_telnet_client_send_tty_naws(bind_host,
                                     unused_tcp_port):
    """Test Client's NAWS of callback method send_naws()."""
    # given a client,
    _waiter = asyncio.Future()
    given_cols, given_rows = 40, 20
    prog, args = 'asynctelnet-client', [
        bind_host, str(unused_tcp_port), '--loglevel=warning',
        '--connect-minwait=0.005', '--connect-maxwait=0.010']

    # a server,
    class ServerTestNaws(asynctelnet.TelnetServer):
        def on_naws(self, width, height):
            super().on_naws(width, height)
            _waiter.set_result((height, width))
            event_loop.call_soon(self.connection_lost, None)

    await asynctelnet.create_server(
        protocol_factory=ServerTestNaws,
        host=bind_host, port=unused_tcp_port,
        connect_maxwait=0.05)

    proc = pexpect.spawn(prog, args, dimensions=(given_rows, given_cols))
    await proc.expect(pexpect.EOF, async_=True, timeout=5)
    assert proc.match == pexpect.EOF

    recv_cols, recv_rows = await asyncio.wait_for(_waiter, 0.5)
    assert recv_cols == given_cols
    assert recv_rows == given_rows


@pytest.mark.anyio
async def test_telnet_client_send_naws_65534(bind_host, unused_tcp_port):
    """Test Client's NAWS boundary values."""
    # given a server
    _waiter = asyncio.Future()
    given_cols, given_rows = 9999999, -999999
    expect_cols, expect_rows = 65535, 0

    class ServerTestNaws(asynctelnet.TelnetServer):
        def on_naws(self, width, height):
            super().on_naws(width, height)
            _waiter.set_result((height, width))

    await asynctelnet.create_server(
        protocol_factory=ServerTestNaws,
        host=bind_host, port=unused_tcp_port,
        connect_maxwait=0.05)

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        cols=given_cols, rows=given_rows, connect_minwait=0.05)

    recv_cols, recv_rows = await asyncio.wait_for(_waiter, 0.5)
    assert recv_cols == expect_cols
    assert recv_rows == expect_rows



"""Accessory functions."""
# std imports
import pkg_resources
import importlib
import logging
import outcome
import anyio

__all__ = ('encoding_from_lang', 'name_unicode', 'eightbits', 'make_logger',
           'repr_mapping', 'function_lookup', 'CtxObj', 'spawn', 'ValueEvent')


def get_version():
    try:
        return pkg_resources.get_distribution("asynctelnet").version
    except Exception:
        return "0.0"


def encoding_from_lang(lang):
    """
    Parse encoding from LANG environment value.

    Example::

        >>> encoding_from_lang('en_US.UTF-8@misc')
        'UTF-8'
    """
    encoding = lang
    if '.' in lang:
        _, encoding = lang.split('.', 1)
    if '@' in encoding:
        encoding, _ = encoding.split('@', 1)
    return encoding


def name_unicode(ucs):
    """Return 7-bit ascii printable of any string. """
    # more or less the same as curses.ascii.unctrl -- but curses
    # module is conditionally excluded from many python distributions!
    bits = ord(ucs)
    if 32 <= bits <= 126:
        # ascii printable as one cell, as-is
        rep = chr(bits)
    elif bits == 127:
        rep = "^?"
    elif bits < 32:
        rep = "^" + chr(((bits & 0x7f) | 0x20) + 0x20)
    else:
        rep = r'\x{:02x}'.format(bits)
    return rep


def eightbits(number):
    """
    Binary representation of ``number`` padded to 8 bits.

    Example::

        >>> eightbits(ord('a'))
        '0b01100001'
    """
    # useful only so far in context of a forwardmask or any bitmask.
    prefix, value = bin(number).split('b')
    return '0b%0.8i' % (int(value),)

_DEFAULT_LOGFMT = ' '.join(('%(asctime)s',
                            '%(levelname)s',
                            '%(filename)s:%(lineno)d',
                            '%(message)s'))
def make_logger(name, loglevel='info', logfile=None, logfmt=_DEFAULT_LOGFMT):
    """Create and return simple logger for the given arguments.
    This is only suitable for your main program.
    """
    lvl = getattr(logging, loglevel.upper())
    logging.getLogger().setLevel(lvl)

    _cfg = {'format': logfmt}
    if logfile:
        _cfg['filename'] = logfile
    logging.basicConfig(**_cfg)
    return logging.getLogger(name)

def repr_mapping(mapping):
    """Return printable string, 'key=value [key=value ...]' for mapping."""
    return ' '.join('='.join(map(str, kv)) for kv in mapping.items())

def function_lookup(pymod_path):
    """Return callable function target from standard module.function path."""
    module_name, func_name = pymod_path.rsplit('.', 1)
    module = importlib.import_module(module_name)
    shell_function = getattr(module, func_name)
    assert callable(shell_function), shell_function
    return shell_function

class CtxObj:
    """
    Add an async context manager that calls `_ctx` to run the context.

    Usage::
        class Foo(CtxObj):
            @asynccontextmanager
            async def _ctx(self):
                yield self  # or whatever

        async with Foo() as self_or_whatever:
            pass
    """
    __ctx = None
    def __aenter__(self):
        if self.__ctx is not None:
            breakpoint()
            raise RuntimeError("Double context")
        self.__ctx = ctx = self._ctx()
        return ctx.__aenter__()

    async def __aexit__(self, *tb):
        ctx,self.__ctx = self.__ctx,None
        if hasattr(self,"aclose"):
            async with anyio.move_on_after(2, shield=True):
                await self.aclose()
        return await ctx.__aexit__(*tb)


async def spawn(tg, proc,*args, _name=None, **kwargs):
    """
    Helper to start a subtask. Like `anyio.abc.TaskGroup.spawn` but
    (a) accepts keyword arguments, (b) returns an `anyio.abc.CancelScope`
    which can be used to kill the task.
    """
    sc = None
    async def _spawn(evt, p,a,k):
        nonlocal sc
        async with anyio.open_cancel_scope() as sc:
            await evt.set()
            await p(*a,**k)

    evt = anyio.create_event()
    await tg.spawn(_spawn, evt, proc,args,kwargs, name=_name)
    await evt.wait()
    return sc

class ValueEvent:
    """A waitable value useful for inter-task synchronization,
    inspired by :class:`threading.Event`.

    An event object manages an internal value, which is initially
    unset, and a task can wait for it to become True.

    Note that the value can only be read once.
    """

    event = None
    value = None

    def __init__(self):
        self.event = anyio.create_event()

    async def set(self, value):
        """Set the result to return this value, and wake any waiting task.
        """
        assert not self.event.is_set(), self
        self.value = outcome.Value(value)
        await self.event.set()

    async def set_error(self, exc):
        """Set the result to raise this exceptio, and wake any waiting task.
        """
        assert not self.event.is_set(), self
        self.value = outcome.Error(exc)
        await self.event.set()

    def is_set(self):
        """Check whether the event has occurred.
        """
        return self.value is not None

    async def get(self):
        """Block until the value is set.

        If it's already set, then this method returns immediately.

        The value can only be read once.
        """
        await self.event.wait()
        return self.value.unwrap()

    async def wait(self):
        """
        Block until the value is set.

        This does not retrieve the value and is meant for locks and similar helpers.
        """
        await self.event.wait()




"""
Special Line Character support for Telnet Linemode Option (:rfc:`1184`).
"""
from enum import IntEnum

from .accessories import eightbits, name_unicode

__all__ = ('SLC', 'NSLC', 'BSD_SLC_TAB', 'generate_slctab', 'Linemode',
           'snoop', 'generate_forwardmask', 'Forwardmask',
           'name_slc_command',
           'NoSupport', 'Var','Flush','Cmd', 'LMode', 'LMode_Mode')

def _exp(cls):
    for k in dir(cls):
        if k[0].isupper():
            globals()[k] = getattr(cls, k)
    return cls

@_exp
class Var(IntEnum):
    NOSUPPORT = 0
    CANTCHANGE = 1
    VARIABLE = 2
    DEFAULT = 3

@_exp
class Flush(IntEnum):
    OUT = 32
    IN = 64
    ACK = 128

SLC_LEVELBITS = 0x03

@_exp
class Cmd(IntEnum):
    DM = 1
    SYNCH = 1  # alias for DM
    BRK = 2
    IP = 3
    AO = 4
    AYT = 5
    EOR = 6
    ABORT = 7
    EOF = 8
    SUSP = 9
    EC = 10
    EL = 11
    EW = 12
    RP = 13
    LNEXT = 14
    XON = 15
    XOFF = 16
    FORW1 = 17
    FORW2 = 18
    MCL = 19
    MCR = 20
    MCWL = 21
    MCWR = 22
    MCBOL = 23
    MCEOL = 24
    INSRT = 25
    OVER = 26
    ECR = 27
    EWR = 28
    EBOL = 29
    EEOL = 30
    NSLC = 31

@_exp
class LMode(IntEnum):
    MODE = 1
    FORWARDMASK = 3
    SLC = 3

@_exp
class LMode_Mode(IntEnum):
    REMOTE = 0
    LOCAL = 1
    TRAPSIG = 2

    # Flags
    ACK = 4
    SOFT_TAB = 8
    LIT_ECHO = 16

class SLC(object):
    def __init__(self, mask=Var.DEFAULT, value=None):
        """
        Defines the willingness to support a Special Linemode Character.

        Defined by its SLC support level, ``mask`` and default keyboard
        ASCII byte ``value`` (may be negotiated by client).
        """
        #   The default byte mask ``Var.DEFAULT`` and value ``b'\x00'`` infer
        #   our willingness to support the option, but with no default value.
        #   The value must be negotiated by client to activate the callback.
        self.mask = mask
        self.val = value

    @property
    def level(self):
        """ Returns SLC level of support.  """
        return self.mask & SLC_LEVELBITS

    @property
    def nosupport(self):
        """ Returns True if SLC level is Var.NOSUPPORT. """
        return self.level == Var.NOSUPPORT

    @property
    def cantchange(self):
        """ Returns True if SLC level is Var.CANTCHANGE. """
        return self.level == Var.CANTCHANGE

    @property
    def variable(self):
        """ Returns True if SLC level is Var.VARIABLE. """
        return self.level == Var.VARIABLE

    @property
    def default(self):
        """ Returns True if SLC level is Var.DEFAULT. """
        return self.level == Var.DEFAULT

    @property
    def ack(self):
        """ Returns True if Flush.ACK bit is set. """
        return self.mask & Flush.ACK

    @property
    def flushin(self):
        """ Returns True if Flush.IN bit is set. """
        return self.mask & Flush.IN

    @property
    def flushout(self):
        """ Returns True if Flush.IN bit is set.  """
        return self.mask & Flush.OUT

    def set_value(self, value):
        """ Set SLC keyboard ascii value to ``byte``.  """
        assert isinstance(mask,int)
        self.val = value

    def set_mask(self, mask):
        """ Set SLC option mask, ``mask``.  """
        assert isinstance(mask,int)
        self.mask = mask

    def set_flag(self, flag):
        """ Set SLC option flag, ``flag``.  """
        assert isinstance(mask,int)
        self.mask = self.mask | flag

    def __str__(self):
        """ SLC definition as string '(value, flag(|s))'. """
        flags = list()
        for flag in ('nosupport', 'variable', 'default', 'ack',
                     'flushin', 'flushout', 'cantchange', ):
            if getattr(self, flag):
                flags.append(flag)
        return '({value}, {flags})'.format(
            value=(name_unicode(self.val)
                   if self.val != _POSIX_VDISABLE
                   else '(DISABLED:\\xff)'),
            flags='|'.join(flags))


class NoSupport(SLC):
    def __init__(self):
        """
        SLC definition inferring our unwillingness to support the option.
        """
        super().__init__(Var.NOSUPPORT, _POSIX_VDISABLE)

#: SLC value may be changed, flushes input and output
_SLC_VARIABLE_FIO = Var.VARIABLE | Flush.IN | Flush.OUT
#: SLC value may be changed, flushes input
_SLC_VARIABLE_FI = Var.VARIABLE | Flush.IN
#: SLC value may be changed, flushes output
_SLC_VARIABLE_FO = Var.VARIABLE | Flush.OUT
#: SLC function for this value is not supported
_POSIX_VDISABLE = b'\xff'

#: This SLC tab when sent to a BSD client warrants no reply; their
#  tabs match exactly. These values are found in ttydefaults.h of
#  termios family of functions.
BSD_SLC_TAB = {
    Cmd.FORW1: NoSupport(),  # unsupported; causes all buffered
    Cmd.FORW2: NoSupport(),  # characters to be sent immediately,
    Cmd.EOF: SLC(Var.VARIABLE,        b'\x04'),  # ^D VEOF
    Cmd.EC: SLC(Var.VARIABLE,         b'\x7f'),  # BS VERASE
    Cmd.EL: SLC(Var.VARIABLE,         b'\x15'),  # ^U VKILL
    Cmd.IP: SLC(_SLC_VARIABLE_FIO,    b'\x03'),  # ^C VINTR
    Cmd.ABORT: SLC(_SLC_VARIABLE_FIO, b'\x1c'),  # ^\ VQUIT
    Cmd.XON: SLC(Var.VARIABLE,        b'\x11'),  # ^Q VSTART
    Cmd.XOFF: SLC(Var.VARIABLE,       b'\x13'),  # ^S VSTOP
    Cmd.EW: SLC(Var.VARIABLE,         b'\x17'),  # ^W VWERASE
    Cmd.RP: SLC(Var.VARIABLE,         b'\x12'),  # ^R VREPRINT
    Cmd.LNEXT: SLC(Var.VARIABLE,      b'\x16'),  # ^V VLNEXT
    Cmd.AO: SLC(_SLC_VARIABLE_FO,     b'\x0f'),  # ^O VDISCARD
    Cmd.SUSP: SLC(_SLC_VARIABLE_FI,   b'\x1a'),  # ^Z VSUSP
    Cmd.AYT: SLC(Var.VARIABLE,        b'\x14'),  # ^T VSTATUS
    # no default value for break, sync, end-of-record,
    Cmd.BRK: SLC(), Cmd.DM: SLC(), Cmd.EOR: SLC(),

}


def generate_slctab(tabset=BSD_SLC_TAB):
    """ Returns full 'SLC Tab' for definitions found using ``tabset``.
        Functions not listed in ``tabset`` are set as Var.NOSUPPORT.
    """
    #   ``slctab`` is a dictionary of SLC functions, such as SLC_IP,
    #   to a tuple of the handling character and support level.
    _slctab = {}
    for slc in [bytes([const]) for const in range(1, NSLC + 1)]:
        _slctab[slc] = tabset.get(slc, NoSupport())
    return _slctab


def generate_forwardmask(binary_mode, tabset, ack=False):
    """
    Generate a :class:`~.Forwardmask` instance.

    Generate a 32-byte (``binary_mode`` is True) or 16-byte (False) Forwardmask
    instance appropriate for the specified ``slctab``.  A Forwardmask is formed
    by a bitmask of all 256 possible 8-bit keyboard ascii input, or, when not
    'outbinary', a 16-byte 7-bit representation of each value, and whether or
    not they should be "forwarded" by the client on the transport stream
    """
    num_bytes, msb = (32, 256) if binary_mode else (16, 127)
    mask32 = []
    for mask in range(msb // 8):
        start = mask * 8
        last = start + 7
        byte = 0
        for char in range(start, last + 1):
            (func, slc_name, slc_def) = snoop(bytes([char]), tabset, dict())
            byte <<= 1
            if func is not None and not slc_def.nosupport:
                # set bit for this character, it is a supported slc char
                byte |= 1
        mask32.append(byte)
    return Forwardmask(bytes(mask32), ack)


def snoop(byte, slctab, caller):
    """ Scan ``slctab`` for matching ``byte`` values.

        Returns (callback, func_byte, slc_definition) on match.
        Otherwise, (None, None, None). If no callback is assigned,
        the value of callback is always None.
    """
    for slc_func, slc_def in slctab.items():
        if byte == slc_def.val and slc_def.val != 0:
            return (caller.get_slc_callback(slc_func), slc_func, slc_def)
    return (None, None, None)


class Linemode(object):
    """ """

    def __init__(self, mask=0):
        """ A mask of ``LMODE_MODE_LOCAL`` means that all line editing is
            performed on the client side (default). A mask of 0
            indicates that editing is performed on the remote side.
            Valid bit flags of mask are: ``LMode_Mode.TRAPSIG``,
            ``LMode_Mode.ACK``, ``LMode_Mode.SOFT_TAB``, and
            ``LMode_Mode.LIT_ECHO``.
        """
        assert isinstance(mask, int), (repr(mask), mask)
        self.mask = mask

    def __eq__(self, other):
        """Compare by another Linemode (LMode_Mode.ACK ignored)."""
        # the inverse OR(|) of acknowledge bit UNSET in comparator,
        # would be the AND OR(& ~) to compare modes without acknowledge
        # bit set.
        return (
            (self.mask | ord(LMode_Mode.ACK)) ==
            (other.mask | ord(LMode_Mode.ACK))
        )

    @property
    def local(self):
        """ True if linemode is local. """
        return bool(self.mask & ord(LMode_Mode.LOCAL))

    @property
    def remote(self):
        """ True if linemode is remote. """
        return not self.local

    @property
    def trapsig(self):
        """ True if signals are trapped by client. """
        return bool(self.mask & ord(LMode_Mode.TRAPSIG))

    @property
    def ack(self):
        """ Returns True if mode has been acknowledged. """
        return bool(self.mask & ord(LMode_Mode.ACK))

    @property
    def soft_tab(self):
        """ Returns True if client will expand horizontal tab (\x09). """
        return bool(self.mask & ord(LMode_Mode.SOFT_TAB))

    @property
    def lit_echo(self):
        """ Returns True if non-printable characters are displayed as-is. """
        return bool(self.mask & ord(LMode_Mode.LIT_ECHO))

    def __str__(self):
        """ Returns string representation of line mode, for debugging """
        return 'remote' if self.remote else 'local'

    def __repr__(self):
        return '<{0!r}: {1}>'.format(
            self.mask, ', '.join([
                '{0}:{1}'.format(prop, getattr(self, prop))
                for prop in ('lit_echo', 'soft_tab', 'ack',
                             'trapsig', 'remote', 'local')])
        )


class Forwardmask(object):
    """ """

    def __init__(self, value, ack=False):
        """
        Forwardmask object using the bytemask value received by server.

        :param bytes value: bytemask ``value`` received by server after ``IAC SB
            LINEMODE DO FORWARDMASK``. It must be a bytearray of length 16 or 32.
        """
        assert isinstance(value, (bytes, bytearray)), value
        assert len(value) in (16, 32), len(value)
        self.value = value
        self.ack = ack

    def description_table(self):
        """
        Returns list of strings describing obj as a tabular ASCII map.
        """
        result = []
        MRK_CONT = '(...)'
        continuing = lambda: len(result) and result[-1] == MRK_CONT
        is_last = lambda mask: mask == len(self.value) - 1
        same_as_last = lambda row: (
            len(result) and result[-1].endswith(row.split()[-1]))

        for mask, byte in enumerate(self.value):
            if byte == 0:
                if continuing() and not is_last(mask):
                    continue
                row = '[%2d] %s' % (mask, eightbits(0),)
                if not same_as_last(row) or is_last(mask):
                    result.append(row)
                else:
                    result.append(MRK_CONT)
            else:
                start = mask * 8
                last = start + 7
                characters = ', '.join([name_unicode(chr(char))
                                        for char in range(start, last + 1)
                                        if char in self])
                result.append('[%2d] %s %s' % (
                    mask, eightbits(byte), characters,))
        return result

    def __str__(self):
        """Returns single string of binary 0 and 1 describing obj."""
        return '0b%s' % (''.join([value for (prefix, value) in [
            eightbits(byte).split('b') for byte in self.value]]),)

    def __contains__(self, number):
        """Whether forwardmask contains keycode ``number``."""
        mask, flag = number // 8, 2 ** (7 - (number % 8))
        return bool(self.value[mask] & flag)


#: List of globals that may match an slc function byte
_DEBUG_SLC_OPTS = dict([(value, key)
                        for key, value in locals().items() if key in
                        ('SLC_SYNCH', 'SLC_BRK', 'SLC_IP', 'SLC_AO', 'SLC_AYT',
                            'SLC_EOR', 'SLC_ABORT', 'SLC_EOF', 'SLC_SUSP',
                            'SLC_EC', 'SLC_EL', 'SLC_EW', 'SLC_RP',
                            'SLC_LNEXT', 'SLC_XON', 'SLC_XOFF', 'SLC_FORW1',
                            'SLC_FORW2', 'SLC_MCL', 'SLC_MCR', 'SLC_MCWL',
                            'SLC_MCWR', 'SLC_MCBOL', 'SLC_MCEOL', 'SLC_INSRT',
                            'SLC_OVER', 'SLC_ECR', 'SLC_EWR', 'SLC_EBOL',
                            'SLC_EEOL',)])


def name_slc_command(byte):
    """ Given an SLC ``byte``, return global mnemonic as string. """
    return (repr(byte) if byte not in _DEBUG_SLC_OPTS
            else _DEBUG_SLC_OPTS[byte])



"""Test instantiation of basic server and client forms."""
# std imports
import asyncio
import os
import sys
import tempfile
import time
import unittest.mock

# local imports
import asynctelnet
from tests.accessories import (
    unused_tcp_port,
    bind_host
)

# 3rd party
import pytest
import pexpect


@pytest.mark.anyio
async def test_create_server(server):
    """Test asynctelnet.create_server basic instantiation."""
    # exercise,
    async with server() as port:
        pass

@pytest.mark.anyio
async def test_open_connection(bind_host, server):
    """Exercise asynctelnet.open_connection with default options."""
    evt = anyio.create_event()
    async def shell(stream):
        await evt.set()
        assert repr(stream) == (
            '<TelnetStream server mode:kludge +lineflow -xon_any +slc_sim '
            'server-will:BINARY,ECHO,SGA '
            'client-will:BINARY,CHARSET,NAWS,NEW_ENVIRON,TTYPE>')
        await anyio.sleep(0.5)

    async with server(shell=shell) as port:
        async with open_connection(host=bind_host, port=port) as client:
            assert repr(client) == (
                '<TelnetStream client mode:kludge +lineflow -xon_any +slc_sim '
                'client-will:BINARY,CHARSET,NAWS,NEW_ENVIRON,TTYPE '
                'server-will:BINARY,ECHO,SGA>')


@pytest.mark.anyio
async def test_create_server_conditionals(server):
    """Test asynctelnet.create_server conditionals."""
    # exercise,
    async with server(
            protocol_factory=lambda: asynctelnet.TelnetServer) as port:
        pass


@pytest.mark.anyio
async def test_create_server_on_connect(bind_host, server):
    """Test on_connect() anonymous function callback of create_server."""
    # given,
    given_pf = unittest.mock.MagicMock()
    async with server(shell=shell, protocol_factory=given_pf) as port:
        async with open_connection(host=bind_host, port=port) as client:
            pass

    # verify
    assert given_pf.called


@pytest.mark.anyio
async def test_telnet_server_open_close(bind_host, server):
    """Test asynctelnet.TelnetServer() instantiation and connection_made()."""
    from asynctelnet.telopt import IAC, WONT, TTYPE
    # given,
    evt = anyio.create_event()
    async def shell(s):
        await evt.set()
        await s.send("Goodbye!")
    async with server(shell=shell) as port:
        async with await anyio.connect_tcp(bind_host, port) as s:
            await s.send(bytes([IAC, WONT, TTYPE]) + b'bye\r')
            await evt.wait()
            result = await s.receive()
            assert result == b'\xff\xfd\x18Goodbye!', result


#@pytest.mark.anyio
#async def test_telnet_client_open_close_by_write(
#        bind_host, unused_tcp_port):
#    """Exercise BaseClient.connection_lost() on writer closed."""
#    await event_loop.create_server(asyncio.Protocol,
#                                        bind_host, unused_tcp_port)
#
#    reader, writer = await asynctelnet.open_connection(
#        host=bind_host, port=unused_tcp_port,
#        connect_minwait=0.05)
#    writer.close()
#    assert (await reader.read()) == ''
#
#
#@pytest.mark.anyio
#async def test_telnet_client_open_closed_by_peer(
#        bind_host, unused_tcp_port):
#    """Exercise BaseClient.connection_lost()."""
#    class DisconnecterProtocol(asyncio.Protocol):
#        def connection_made(self, transport):
#            # disconnect on connect
#            transport.close()
#
#    await event_loop.create_server(DisconnecterProtocol,
#                                        bind_host, unused_tcp_port)
#
#    reader, writer = await asynctelnet.open_connection(
#        host=bind_host, port=unused_tcp_port,
#        connect_minwait=0.05)
#
#    # read until EOF, no data received.
#    data_received = await reader.read()
#    assert data_received == ''
#

@pytest.mark.anyio
async def test_telnet_server_advanced_negotiation(
        bind_host, unused_tcp_port):
    """Test asynctelnet.TelnetServer() advanced negotiation."""
    # given
    from asynctelnet.telopt import (
        IAC, DO, WILL, SB, TTYPE, NEW_ENVIRON, NAWS, SGA, ECHO, CHARSET, BINARY
    )
    _waiter = asyncio.Future()

    class ServerTestAdvanced(asynctelnet.TelnetServer):
        def begin_advanced_negotiation(self):
            super().begin_advanced_negotiation()
            _waiter.set_result(self)

    await asynctelnet.create_server(
        protocol_factory=ServerTestAdvanced,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    writer.write(IAC + WILL + TTYPE)
    server = await asyncio.wait_for(_waiter, 0.5)

    # verify,
    assert server.writer.remote_option[TTYPE] is True
    assert server.writer.pending_option == {
        # server's request to negotiation TTYPE affirmed
        DO + TTYPE: False,
        # server's request for TTYPE value unreplied
        SB + TTYPE: True,
        # remaining unreplied values from begin_advanced_negotiation()
        DO + NEW_ENVIRON: True,
        DO + CHARSET: True,
        DO + NAWS: True,
        WILL + SGA: True,
        WILL + ECHO: True,
        WILL + BINARY: True,
    }


@pytest.mark.anyio
async def test_telnet_server_closed_by_client(
        bind_host, unused_tcp_port):
    """Exercise TelnetServer.connection_lost."""
    # given
    _waiter = asyncio.Future()

    await asynctelnet.create_server(
        _waiter_closed=_waiter,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    writer.close()

    # verify,
    srv_instance = await asyncio.wait_for(_waiter, 0.5)
    assert srv_instance._closing

    # exercise, a 2nd call to .connection_lost() returns early,
    # allowing callbacks the freedom to call it at any time from
    # the server-end to dump the client.
    srv_instance.connection_lost(exc=None)


@pytest.mark.anyio
async def test_telnet_server_eof_by_client(
        bind_host, unused_tcp_port):
    """Exercise TelnetServer.eof_received()."""
    # given
    _waiter = asyncio.Future()

    await asynctelnet.create_server(
        _waiter_closed=_waiter,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    writer.write_eof()

    # verify,
    srv_instance = await asyncio.wait_for(_waiter, 0.5)
    assert srv_instance._closing


@pytest.mark.anyio
async def test_telnet_server_closed_by_server(
        bind_host, unused_tcp_port):
    """Exercise TelnetServer.connection_lost by close()."""
    from asynctelnet.telopt import IAC, DO, WONT, TTYPE

    # given
    _waiter_connected = asyncio.Future()
    _waiter_closed = asyncio.Future()

    await asynctelnet.create_server(
        _waiter_connected=_waiter_connected,
        _waiter_closed=_waiter_closed,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # data received by client, connection is made
    expect_hello = IAC + DO + TTYPE
    hello_reply = IAC + WONT + TTYPE + b'quit' + b'\r\n'

    # exercise,
    hello = await reader.readexactly(len(expect_hello))

    # verify,
    assert hello == expect_hello

    # exercise,
    writer.write(hello_reply)
    server = await asyncio.wait_for(_waiter_connected, 0.5)

    # exercise, by closing.
    server.writer.close()

    # verify
    await asyncio.wait_for(_waiter_closed, 0.5)


@pytest.mark.anyio
async def test_telnet_server_idle_duration(
        bind_host, unused_tcp_port):
    """Exercise TelnetServer.idle property."""
    from asynctelnet.telopt import IAC, WONT, TTYPE

    # given
    _waiter_connected = asyncio.Future()
    _waiter_closed = asyncio.Future()

    await asynctelnet.create_server(
        _waiter_connected=_waiter_connected,
        _waiter_closed=_waiter_closed,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    writer.write(IAC + WONT + TTYPE)
    server = await asyncio.wait_for(_waiter_connected, 0.5)

    # verify
    assert 0 <= server.idle <= 0.5
    assert 0 <= server.duration <= 0.5


@pytest.mark.anyio
async def test_telnet_client_idle_duration_minwait(
        bind_host, unused_tcp_port):
    """Exercise TelnetClient.idle property and minimum connection time."""
    from asynctelnet.telopt import IAC, WONT, TTYPE

    # a server that doesn't care
    await event_loop.create_server(asyncio.Protocol,
                                        bind_host, unused_tcp_port)

    given_minwait = 0.100

    stime = time.time()
    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        connect_minwait=given_minwait, connect_maxwait=given_minwait)

    elapsed_ms = int((time.time() - stime) * 1e3)
    expected_ms = int(given_minwait * 1e3)
    assert expected_ms <= elapsed_ms <= expected_ms + 50

    # verify
    assert 0 <= writer.protocol.idle <= 0.5
    assert 0 <= writer.protocol.duration <= 0.5


@pytest.mark.anyio
async def test_telnet_server_closed_by_error(
        bind_host, unused_tcp_port):
    """Exercise TelnetServer.connection_lost by exception."""
    from asynctelnet.telopt import IAC, DO, WONT, TTYPE

    # given
    _waiter_connected = asyncio.Future()
    _waiter_closed = asyncio.Future()

    await asynctelnet.create_server(
        _waiter_connected=_waiter_connected,
        _waiter_closed=_waiter_closed,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    writer.write(IAC + WONT + TTYPE)
    server = await asyncio.wait_for(_waiter_connected, 0.5)

    class CustomException(Exception):
        pass

    # exercise, by connection_lost(exc=Exception())..
    server.writer.write('Bye!')
    server.connection_lost(CustomException('blah!'))

    # verify, custom exception is thrown into any yielding reader
    with pytest.raises(CustomException):
        await server.reader.read()


@pytest.mark.anyio
async def test_telnet_client_open_close_by_error(
        bind_host, unused_tcp_port):
    """Exercise BaseClient.connection_lost() on error."""
    await event_loop.create_server(asyncio.Protocol,
                                        bind_host, unused_tcp_port)

    class GivenException(Exception):
        pass

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port, connect_minwait=0.05)

    writer.protocol.connection_lost(GivenException("candy corn 4 everyone"))
    with pytest.raises(GivenException):
        await reader.read()


@pytest.mark.anyio
async def test_telnet_server_negotiation_fail(
        bind_host, unused_tcp_port):
    """Test asynctelnet.TelnetServer() negotiation failure with client."""
    from asynctelnet.telopt import DO, TTYPE
    # given
    _waiter_connected = asyncio.Future()

    await asynctelnet.create_server(
        _waiter_connected=_waiter_connected,
        host=bind_host, port=unused_tcp_port,
        connect_maxwait=0.05)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    await reader.readexactly(3)  # IAC DO TTYPE, we ignore it!

    # negotiation then times out, deferring to waiter_connected.
    server = await asyncio.wait_for(_waiter_connected, 1.0)

    # verify,
    assert server.negotiation_should_advance() is False
    assert server.writer.pending_option[DO + TTYPE] == True

    assert repr(server.writer) == ('<TelnetWriter server '
                                   'mode:local +lineflow -xon_any +slc_sim '
                                   'failed-reply:DO TTYPE>')

@pytest.mark.anyio
async def test_telnet_client_negotiation_fail(
        bind_host, unused_tcp_port):
    """Test asynctelnet.TelnetCLient() negotiation failure with server."""

    class ClientNegotiationFail(asynctelnet.TelnetClient):
        def connection_made(self, transport):
            from asynctelnet.telopt import WILL, TTYPE
            super().connection_made(transport)
            # this creates a pending negotiation demand from the client-side.
            self.writer.iac(WILL, TTYPE)

    # a server that never responds with nothing.
    await event_loop.create_server(asyncio.Protocol,
                                        bind_host, unused_tcp_port)

    given_minwait = 0.05
    given_maxwait = 0.100

    stime = time.time()
    reader, writer = await asyncio.wait_for(asynctelnet.open_connection(
        client_factory=ClientNegotiationFail, host=bind_host,
        port=unused_tcp_port,
        connect_minwait=given_minwait,
        connect_maxwait=given_maxwait), 5)

    elapsed_ms = int((time.time() - stime) * 1e3)
    expected_ms = int(given_maxwait * 1e3)
    assert expected_ms <= elapsed_ms <= expected_ms + 50


@pytest.mark.anyio
async def test_telnet_server_as_module():
    """Test __main__ hook, when executing python -m asynctelnet.server --help"""
    prog = sys.executable
    args = [prog, '-m', 'asynctelnet.server', '--help']
    proc = await asyncio.create_subprocess_exec(
        *args, stdout=asyncio.subprocess.PIPE,
               stderr=asyncio.subprocess.PIPE)

    # we would expect the script to display help output and exit
    help_output, _ = await proc.communicate()
    assert help_output.startswith(b'usage: server.py [-h]')
    await proc.communicate()
    await proc.wait()


@pytest.mark.anyio
async def test_telnet_server_cmdline(bind_host, unused_tcp_port):
    """Test executing asynctelnet/server.py as server"""
    # this code may be reduced when pexpect asyncio is bugfixed ..
    prog = pexpect.which('asynctelnet-server')
    args = [prog, bind_host, str(unused_tcp_port), '--loglevel=info',
            '--connect-maxwait=0.05']
    proc = await asyncio.create_subprocess_exec(
        *args, stderr=asyncio.subprocess.PIPE)

    seen = b''
    while True:
        line = await asyncio.wait_for(proc.stderr.readline(), 0.5)
        if b'Server ready' in line:
            break
        seen += line
        assert line, seen.decode()  # EOF reached

    # client connects,
    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # and closes,
    await reader.readexactly(3)  # IAC DO TTYPE, we ignore it!
    writer.close()

    seen = b''
    while True:
        line = await asyncio.wait_for(proc.stderr.readline(), 0.5)
        if b'Connection closed' in line:
            break
        seen += line
        assert line, seen.decode()  # EOF reached

    # send SIGTERM
    proc.terminate()

    # we would expect the server to gracefully end.
    await proc.communicate()
    await proc.wait()


@pytest.mark.anyio
async def test_telnet_client_as_module():
    """Test __main__ hook, when executing python -m asynctelnet.client --help"""
    prog = sys.executable
    args = [prog, '-m', 'asynctelnet.client', '--help']
    proc = await asyncio.create_subprocess_exec(
        *args, stdout=asyncio.subprocess.PIPE,
               stderr=asyncio.subprocess.PIPE)

    # we would expect the script to display help output and exit
    help_output, _ = await proc.communicate()
    assert help_output.startswith(b'usage: client.py [-h]')
    await proc.communicate()
    await proc.wait()


@pytest.mark.anyio
async def test_telnet_client_cmdline(bind_host, unused_tcp_port):
    """Test executing asynctelnet/client.py as client"""
    # this code may be reduced when pexpect asyncio is bugfixed ..
    # we especially need pexpect to pass sys.stdin.isatty() test.
    prog = pexpect.which('asynctelnet-client')
    args = [prog, bind_host, str(unused_tcp_port), '--loglevel=info',
            '--connect-minwait=0.05', '--connect-maxwait=0.05']

    class HelloServer(asyncio.Protocol):
        def connection_made(self, transport):
            super().connection_made(transport)
            transport.write(b'hello, space cadet.\r\n')
            # hangup
            event_loop.call_soon(transport.close)

    # start vanilla tcp server
    await event_loop.create_server(HelloServer,
                                        bind_host, unused_tcp_port)

    proc = await asyncio.create_subprocess_exec(
        *args,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE)

    line = await asyncio.wait_for(proc.stdout.readline(), 1.5)
    assert line.strip() == b"Escape character is '^]'."

    line = await asyncio.wait_for(proc.stdout.readline(), 1.5)
    assert line.strip() == b'hello, space cadet.'

    # message received, expect the client to gracefully quit.
    out, err = await asyncio.wait_for(proc.communicate(), 1)
    assert out == b'\x1b[m\nConnection closed by foreign host.\n'


@pytest.mark.anyio
async def test_telnet_client_tty_cmdline(bind_host, unused_tcp_port):
    """Test executing asynctelnet/client.py as client using a tty (pexpect)"""
    # this code may be reduced when pexpect asyncio is bugfixed ..
    # we especially need pexpect to pass sys.stdin.isatty() test.
    prog, args = 'asynctelnet-client', [
        bind_host, str(unused_tcp_port), '--loglevel=warning',
        '--connect-minwait=0.05', '--connect-maxwait=0.05']

    class HelloServer(asyncio.Protocol):
        def connection_made(self, transport):
            super().connection_made(transport)
            transport.write(b'hello, space cadet.\r\n')
            event_loop.call_soon(transport.close)

    # start vanilla tcp server
    await event_loop.create_server(HelloServer,
                                        bind_host, unused_tcp_port)
    proc = pexpect.spawn(prog, args)
    await proc.expect(pexpect.EOF, async_=True, timeout=5)
    # our 'space cadet' has \r\n hardcoded, so \r\r\n happens, ignore it
    assert proc.before == (b"Escape character is '^]'.\r\n"
                           b"hello, space cadet.\r\r\n"
                           b"\x1b[m\r\n"
                           b"Connection closed by foreign host.\r\n")

@pytest.mark.anyio
async def test_telnet_client_cmdline_stdin_pipe(bind_host, unused_tcp_port):
    """Test sending data through command-line client (by os PIPE)."""
    # this code may be reduced when pexpect asyncio is bugfixed ..
    # we especially need pexpect to pass sys.stdin.isatty() test.
    prog = pexpect.which('asynctelnet-client')
    fd, logfile = tempfile.mkstemp(prefix='asynctelnet', suffix='.log')
    os.close(fd)

    args = [prog, bind_host, str(unused_tcp_port), '--loglevel=info',
            '--connect-minwait=0.15', '--connect-maxwait=0.15',
            '--logfile={0}'.format(logfile)]

    async def shell(reader, writer):
        writer.write('Press Return to continue:')
        inp = await reader.readline()
        if inp:
            writer.echo(inp)
            writer.write('\ngoodbye.\n')
        await writer.drain()
        writer.close()

    # start server
    await asynctelnet.create_server(
        host=bind_host, port=unused_tcp_port,
        shell=shell,
        connect_maxwait=0.05)

    # start client by way of pipe
    proc = await asyncio.create_subprocess_exec(
        *args,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    #line = await asyncio.wait_for(proc.stdout.readline(), 1.5)
    #assert line.strip() == b"Escape character is '^]'."

    # message received, expect the client to gracefully quit.
    stdout, stderr = await asyncio.wait_for(proc.communicate(b'\r'), 2)

    # And finally, analyze the contents of the logfile,
    # - 2016-03-18 20:19:25,227 INFO client_base.py:113 Connected to <Peer 127.0.0.1 51237>
    # - 2016-03-18 20:19:25,286 INFO client_base.py:67 Connection closed to <Peer 127.0.0.1 51237>
    logfile_output = open(logfile).read().splitlines()
    assert stdout == (b"Escape character is '^]'.\n"
                      b"Press Return to continue:\r\ngoodbye.\n"
                      b"\x1b[m\nConnection closed by foreign host.\n")

    # verify,
    assert len(logfile_output) == 2, logfile
    assert 'Connected to <Peer' in logfile_output[0], logfile
    assert 'Connection closed to <Peer' in logfile_output[1], logfile
    os.unlink(logfile)



# std imports
import collections
import contextlib
import logging
import codecs
import sys

# local
from . import accessories

__all__ = ('telnet_client_shell', )


# TODO: needs 'wait_for' implementation (see DESIGN.rst)
# task = telnet_writer.wait_for(lambda: telnet_writer.local_mode[ECHO] == True)

if sys.platform == 'win32':
    raise NotImplementedError(
        'win32 not yet supported as telnet client. Please contribute!')

import termios
import os
import anyio
import fcntl

from .stream import ReadCallback
from .telopt import WILL,WONT, ECHO

class TerminalStream(anyio.abc.ByteStream):
    """
    Context manager that yields a non-blocking stdin+stdout stream.

    When sys.stdin is a attached to a terminal, it is configured for
    the telnet mode negotiated. The caller is responsible for setting
    these as appropriate.

    This class *must* be used as a context manager.
    It will attach to stdin/stdout and modify TTY raw mode as appropriate.
    """
    _orig_fl = None

    _ModeDef = collections.namedtuple(
        'mode', ['iflag', 'oflag', 'cflag', 'lflag',
                    'ispeed', 'ospeed', 'cc'])

    charset = "utf-8"
    _decoder = codecs.getincrementaldecoder(charset)()
    _encoder = "utf-8"
    keyboard_escape = b'\x1d'

    def __init__(self, will_echo=True, keyboard_escape=None):
        self._fileno = sys.stdin.fileno()
        self._istty = os.path.sameopenfile(0, 1)
        self._will_echo = will_echo
        if keyboard_escape is not None:
            self.keyboard_escape = keyboard_escape

    async def __aenter__(self):
        await super().__aenter__()
        if self._istty:
            self._saved_mode = self._ModeDef(*termios.tcgetattr(self._fileno))
            self._set_mode()
            self._orig_fl = fcntl.fcntl(self._fileno, fcntl.F_GETFL)
            fcntl.fcntl(self._fileno, fcntl.F_SETFL, self._orig_fl | os.O_NONBLOCK)
        return self

    async def __aexit__(self, *tb):
        self._close()
        await super().__aexit__(*tb)

    def _close(self):
        if self._istty and self._orig_fl is not None:
            fcntl.fcntl(self._fileno, fcntl.F_SETFL, self._orig_fl)
            termios.tcsetattr(
                self._fileno, termios.TCSAFLUSH, list(self._saved_mode))
            self._orig_fl = None

    async def aclose(self):
        self._close()

    async def send_eof():
        pass

    def _set_mode(self):
        termios.tcsetattr(
            sys.stdin.fileno(), termios.TCSAFLUSH, list(self._wanted_mode))

    @property
    def will_echo(self):
        return self._will_echo

    @will_echo.setter
    def will_echo(self, value):
        self._will_echo = value
        if self._istty:
            self._set_mode()

    def fileno(self):
        return self._fileno

    @property
    def _wanted_mode(self):
        """
        Return copy of 'mode' with changes suggested for telnet connection.
        """
        from asynctelnet.telopt import ECHO

        mode = self._saved_mode
        if not self._will_echo:
            # return mode as-is
            return mode

        # "Raw mode", see tty.py function setraw.  This allows sending
        # of ^J, ^C, ^S, ^\, and others, which might otherwise
        # interrupt with signals or map to another character.  We also
        # trust the remote server to manage CR/LF without mapping.
        #
        iflag = mode.iflag & ~(
            termios.BRKINT |  # Do not send INTR signal on break
            termios.ICRNL  |  # Do not map CR to NL on input
            termios.INPCK  |  # Disable input parity checking
            termios.ISTRIP |  # Do not strip input characters to 7 bits
            termios.IXON)     # Disable START/STOP output control

        # Disable parity generation and detection,
        # Select eight bits per byte character size.
        cflag = mode.cflag & ~(termios.CSIZE | termios.PARENB)
        cflag = cflag | termios.CS8

        # Disable canonical input (^H and ^C processing),
        # disable any other special control characters,
        # disable checking for INTR, QUIT, and SUSP input.
        lflag = mode.lflag & ~(
            termios.ICANON | termios.IEXTEN | termios.ISIG | termios.ECHO)

        # Disable post-output processing,
        # such as mapping LF('\n') to CRLF('\r\n') in output.
        oflag = mode.oflag & ~(termios.OPOST | termios.ONLCR)

        # "A pending read is not satisfied until MIN bytes are received
        #  (i.e., the pending read until MIN bytes are received), or a
        #  signal is received.  A program that uses this case to read
        #  record-based terminal I/O may block indefinitely in the read
        #  operation."
        cc = list(mode.cc)
        cc[termios.VMIN] = 1
        cc[termios.VTIME] = 0

        return self._ModeDef(
            iflag=iflag, oflag=oflag, cflag=cflag, lflag=lflag,
            ispeed=mode.ispeed, ospeed=mode.ospeed, cc=cc)

    async def receive(self, max_bytes=1024):
        while True:
            await anyio.wait_socket_readable(self)
            inp = os.read(self._fileno, max_bytes)
            if self.keyboard_escape in inp:
                # on ^], close connection to remote host
                raise EOFError
            inp = self._decoder.decode(inp)
            if inp:
                return inp

    async def send(self, item):
        if isinstance(item,str):
            item = item.encode(self.charset)
        while len(item):
            await anyio.wait_socket_writable(self)
            done = os.write(self._fileno, item)
            item = item[done:]


async def telnet_client_shell(telnet_stream):
    """
    Minimal telnet client shell for POSIX terminals.

    This shell performs minimal tty mode handling when a terminal is
    attached to standard in (keyboard), notably raw mode is often set
    and this shell may exit only by disconnect from server, or the
    escape character, ^].

    stdin or stdout may also be a pipe or file, behaving much like nc(1).

    """

    class TerminalUpdater(ReadCallback):
        def __init__(self,val):
            super().__init__()
            self.val=val
        async def run(self):
            term.will_echo = self.val

    async def _handle_will_echo():
        telnet_stream.queue_read_callback(TerminalUpdater(True))
        return True

    async def _handle_wont_echo():
        telnet_stream.queue_read_callback(TerminalUpdater(False))
        return False

    async with anyio.create_task_group() as tg, \
            TerminalStream(will_echo=telnet_stream.will_echo) as term :
        telnet_stream.set_command_handler(WILL,ECHO, _handle_will_echo)
        telnet_stream.set_command_handler(WONT,ECHO, _handle_wont_echo)

        linesep = '\n'
        if term._istty and term.will_echo:
            linesep = '\r\n'
        await term.send("Escape character is '{escape}'.{linesep}".format(
            escape=accessories.name_unicode(term.keyboard_escape),
            linesep=linesep).encode())

        async def read_stdin():
            while True:
                inp = await term.receive()
                if not inp:
                    break
                if inp:
                    inp = inp.replace('\n','\r\n').replace('\r\r\n','\r\n')
                await telnet_stream.send(inp)

        async def read_telnet():
            while True:
                try:
                    out = await telnet_stream.receive()
                except anyio.EndOfStream:
                    await term.send(f"\033[m{linesep}Connection closed by foreign host.{linesep}".encode())
                    await tg.cancel_scope.cancel()
                    return
                except anyio.ClosedResourceError: 
                    return
                else:
                    await term.send(out)

        await tg.spawn(read_stdin)
        await tg.spawn(read_telnet)



"""Test accessories for asynctelnet project."""
import pytest
import contextlib
from functools import partial
import anyio

from asynctelnet.server import server_loop

@pytest.fixture(scope="module", params=['127.0.0.1'])
def bind_host(request):
    """ Localhost bind address. """
    return request.param

def _unused_tcp_port():
    """Find an unused localhost TCP port from 1024-65535 and return it."""
    with contextlib.closing(socket.socket()) as sock:
        sock.bind(('127.0.0.1', 0))
        return sock.getsockname()[1]

@pytest.fixture(scope="module", params=['127.0.0.1'])
def server(bind_host, unused_tcp_port):
    @asynccontextmanager
    async def mgr(**kw):
        async with anyio.create_task_group() as tg:
            await tg.spawn(partial(server_loop,host=bind_host,port=unused_tcp_port, **kw))
            yield unused_tcp_port
            await tg.cancel_scope.cancel()
    return mgr

@pytest.fixture
def unused_tcp_port():
    return _unused_tcp_port()

@pytest.fixture
def unused_tcp_port_factory():
    """A factory function, producing different unused TCP ports."""
    produced = set()

    def factory():
        """Return an unused port."""
        port = _unused_tcp_port()

        while port in produced:
            port = _unused_tcp_port()

        produced.add(port)

        return port
    return factory


__all__ = ('bind_host', 'unused_tcp_port', 'unused_tcp_port_factory',)



"""Test XDISPLOC, rfc-1096_."""
# std imports
import asyncio

# local imports
import asynctelnet
from asynctelnet.tests.accessories import (
    unused_tcp_port,
    bind_host
)

# 3rd party
import pytest


@pytest.mark.anyio
async def test_telnet_server_on_xdisploc(
        bind_host, unused_tcp_port):
    """Test Server's callback method on_xdisploc()."""
    # given
    from asynctelnet.telopt import (
        IAC, WILL, SB, SE, IS, XDISPLOC
    )
    _waiter = asyncio.Future()
    given_xdisploc = 'alpha:0'

    class ServerTestXdisploc(asynctelnet.TelnetServer):
        def on_xdisploc(self, xdisploc):
            super().on_xdisploc(xdisploc)
            _waiter.set_result(self)

    await asynctelnet.create_server(
        protocol_factory=ServerTestXdisploc,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    writer.write(IAC + WILL + XDISPLOC)
    writer.write(IAC + SB + XDISPLOC + IS +
                 given_xdisploc.encode('ascii') +
                 IAC + SE)

    # verify,
    srv_instance = await asyncio.wait_for(_waiter, 0.5)
    assert srv_instance.get_extra_info('xdisploc') == 'alpha:0'


@pytest.mark.anyio
async def test_telnet_client_send_xdisploc(bind_host, unused_tcp_port):
    """Test Client's callback method send_xdisploc()."""
    # given
    _waiter = asyncio.Future()
    given_xdisploc = 'alpha'

    class ServerTestXdisploc(asynctelnet.TelnetServer):
        def on_xdisploc(self, xdisploc):
            super().on_xdisploc(xdisploc)
            _waiter.set_result(xdisploc)

        def begin_advanced_negotiation(self):
            from asynctelnet.telopt import DO, XDISPLOC
            super().begin_advanced_negotiation()
            self.writer.iac(DO, XDISPLOC)

    await asynctelnet.create_server(
        protocol_factory=ServerTestXdisploc,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        xdisploc=given_xdisploc, connect_minwait=0.05)

    recv_xdisploc = await asyncio.wait_for(_waiter, 0.5)
    assert recv_xdisploc == given_xdisploc



"""Test the server's shell(reader, writer) callback."""
# std imports
import asyncio
import time

# local imports
import asynctelnet
from asynctelnet.tests.accessories import (
    unused_tcp_port,
    bind_host
)

# 3rd party
import pytest


@pytest.mark.anyio
async def test_telnet_server_default_timeout(
        bind_host, unused_tcp_port):
    """Test callback on_timeout() as coroutine of create_server()."""
    from asynctelnet.telopt import IAC, WONT, TTYPE
    # given,
    _waiter = asyncio.Future()
    given_timeout = 19.29

    await asynctelnet.create_server(
        _waiter_connected=_waiter,
        host=bind_host, port=unused_tcp_port,
        timeout=given_timeout)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    writer.write(IAC + WONT + TTYPE)

    server = await asyncio.wait_for(_waiter, 0.5)
    assert server.get_extra_info('timeout') == given_timeout

    # exercise, calling set_timeout remains the default given_value.
    server.set_timeout()
    assert server.get_extra_info('timeout') == given_timeout


@pytest.mark.anyio
async def test_telnet_server_set_timeout(
        bind_host, unused_tcp_port):
    """Test callback on_timeout() as coroutine of create_server()."""
    from asynctelnet.telopt import IAC, WONT, TTYPE
    # given,
    _waiter = asyncio.Future()

    # exercise,
    await asynctelnet.create_server(
        _waiter_connected=_waiter,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    writer.write(IAC + WONT + TTYPE)

    server = await asyncio.wait_for(_waiter, 0.5)
    for value in (19.29, 0):
        server.set_timeout(value)
        assert server.get_extra_info('timeout') == value

    # calling with no arguments does nothing, only resets
    # the timer. value remains the last-most value from
    # previous loop
    server.set_timeout()
    assert server.get_extra_info('timeout') == 0


@pytest.mark.anyio
async def test_telnet_server_waitfor_timeout(
        bind_host, unused_tcp_port):
    """Test callback on_timeout() as coroutine of create_server()."""
    from asynctelnet.telopt import IAC, DO, WONT, TTYPE
    # given,
    expected_output = IAC + DO + TTYPE + b'\r\nTimeout.\r\n'

    await asynctelnet.create_server(
        host=bind_host, port=unused_tcp_port,
        timeout=0.050)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    writer.write(IAC + WONT + TTYPE)

    stime = time.time()
    output = await asyncio.wait_for(reader.read(), 0.5)
    elapsed = time.time() - stime
    assert 0.050 <= round(elapsed, 3) <= 0.100
    assert output == expected_output



"""Test XDISPLOC, rfc-1096_."""
# std imports
import asyncio

# local imports
import asynctelnet
from asynctelnet.tests.accessories import (
    unused_tcp_port,
    bind_host
)

# 3rd party
import pytest


@pytest.mark.anyio
async def test_telnet_server_on_charset(
        bind_host, unused_tcp_port):
    """Test Server's callback method on_charset()."""
    # given
    from asynctelnet.telopt import (
        IAC, WILL, WONT, SB, SE, TTYPE, CHARSET, ACCEPTED
    )
    _waiter = asyncio.Future()
    given_charset = 'KOI8-U'

    class ServerTestCharset(asynctelnet.TelnetServer):
        def on_charset(self, charset):
            super().on_charset(charset)
            _waiter.set_result(self)

    await asynctelnet.create_server(
        protocol_factory=ServerTestCharset,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    val = await asyncio.wait_for(reader.readexactly(3), 0.5)
    # exercise,
    writer.write(IAC + WILL + CHARSET)
    writer.write(IAC + WONT + TTYPE)
    writer.write(IAC + SB + CHARSET + ACCEPTED +
                 given_charset.encode('ascii') +
                 IAC + SE)

    # verify,
    srv_instance = await asyncio.wait_for(_waiter, 2.0)
    assert srv_instance.get_extra_info('charset') == given_charset


@pytest.mark.anyio
async def test_telnet_client_send_charset(bind_host, unused_tcp_port):
    """Test Client's callback method send_charset() selection for illegals."""
    # given
    _waiter = asyncio.Future()

    class ServerTestCharset(asynctelnet.TelnetServer):
        def on_request_charset(self):
            return ['illegal', 'cp437']

    class ClientTestCharset(asynctelnet.TelnetClient):
        def send_charset(self, offered):
            selected = super().send_charset(offered)
            _waiter.set_result(selected)
            return selected

    await asyncio.wait_for(
        asynctelnet.create_server(
            protocol_factory=ServerTestCharset,
            host=bind_host, port=unused_tcp_port)
        0.15)

    reader, writer = await asyncio.wait_for(
        asynctelnet.open_connection(
            client_factory=ClientTestCharset,
            host=bind_host, port=unused_tcp_port,
            encoding='latin1', connect_minwait=0.05),
        0.15)

    val = await asyncio.wait_for(_waiter, 1.5)
    assert val == 'cp437'
    assert writer.get_extra_info('charset') == 'cp437'


@pytest.mark.anyio
async def test_telnet_client_no_charset(bind_host, unused_tcp_port):
    """Test Client's callback method send_charset() does not select."""
    # given
    _waiter = asyncio.Future()

    class ServerTestCharset(asynctelnet.TelnetServer):
        def on_request_charset(self):
            return ['illegal', 'this-is-no-good-either']

    class ClientTestCharset(asynctelnet.TelnetClient):
        def send_charset(self, offered):
            selected = super().send_charset(offered)
            _waiter.set_result(selected)
            return selected

    await asynctelnet.create_server(
        protocol_factory=ServerTestCharset,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asynctelnet.open_connection(
        client_factory=ClientTestCharset,
        host=bind_host, port=unused_tcp_port,
        encoding='latin1', connect_minwait=0.05)

    # charset remains latin1
    val = await asyncio.wait_for(_waiter, 0.5)
    assert val == ''
    assert writer.get_extra_info('charset') == 'latin1'



"""Test NEW_ENVIRON, rfc-1572_."""
# std imports
import asyncio

# local imports
import asynctelnet
import asynctelnet.stream
from asynctelnet.tests.accessories import (
    unused_tcp_port,
    bind_host
)

# 3rd party
import pytest


@pytest.mark.anyio
async def test_telnet_server_on_environ(
        bind_host, unused_tcp_port):
    """Test Server's callback method on_environ()."""
    # given
    from asynctelnet.telopt import (
        IAC, WILL, SB, SE, IS, NEW_ENVIRON
    )
    _waiter = asyncio.Future()

    class ServerTestEnviron(asynctelnet.TelnetServer):
        def on_environ(self, mapping):
            super().on_environ(mapping)
            _waiter.set_result(self)

    await asynctelnet.create_server(
        protocol_factory=ServerTestEnviron,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    writer.write(IAC + WILL + NEW_ENVIRON)
    writer.write(IAC + SB + NEW_ENVIRON + IS +
                 asynctelnet.stream._encode_env_buf({
                     # note how the default implementation .upper() cases
                     # all environment keys.
                     'aLpHa': 'oMeGa',
                     'beta': 'b',
                     'gamma': u''.join(chr(n) for n in range(0, 128)),
                 }) + IAC + SE)

    srv_instance = await asyncio.wait_for(_waiter, 0.5)
    assert srv_instance.get_extra_info('ALPHA') == 'oMeGa'
    assert srv_instance.get_extra_info('BETA') == 'b'
    assert srv_instance.get_extra_info('GAMMA') == (
        u''.join(chr(n) for n in range(0, 128)))


@pytest.mark.anyio
async def test_telnet_client_send_environ(bind_host,
                                    unused_tcp_port):
    """Test Client's callback method send_environ() for specific requests."""
    # given
    _waiter = asyncio.Future()
    given_cols = 19
    given_rows = 84
    given_encoding = 'cp437'
    given_term = 'vt220'

    class ServerTestEnviron(asynctelnet.TelnetServer):
        def on_environ(self, mapping):
            super().on_environ(mapping)
            _waiter.set_result(mapping)

    await asynctelnet.create_server(
        protocol_factory=ServerTestEnviron,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        cols=given_cols, rows=given_rows, encoding=given_encoding,
        term=given_term, connect_minwait=0.05)

    mapping = await asyncio.wait_for(_waiter, 0.5)
    assert mapping == {
        'COLUMNS': str(given_cols),
        'LANG': 'en_US.' + given_encoding,
        'LINES': str(given_rows),
        'TERM': 'vt220'
    }


@pytest.mark.anyio
async def test_telnet_client_send_var_uservar_environ(bind_host,
                                                unused_tcp_port):
    """Test Client's callback method send_environ() for VAR/USERVAR request."""
    # given
    _waiter = asyncio.Future()
    given_cols = 19
    given_rows = 84
    given_encoding = 'cp437'
    given_term = 'vt220'

    class ServerTestEnviron(asynctelnet.TelnetServer):
        def on_environ(self, mapping):
            super().on_environ(mapping)
            _waiter.set_result(mapping)

        def on_request_environ(self):
            from asynctelnet.telopt import VAR, USERVAR
            return [VAR, USERVAR]

    await asynctelnet.create_server(
        protocol_factory=ServerTestEnviron,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        cols=given_cols, rows=given_rows, encoding=given_encoding,
        term=given_term, connect_minwait=0.05, connect_maxwait=0.05)

    mapping = await asyncio.wait_for(_waiter, 0.5)
    # although nothing was demanded by server,
    assert mapping == {}

    # the client still volunteered these basic variables,
    mapping == {
        'COLUMNS': str(given_cols),
        'LANG': 'en_US.' + given_encoding,
        'LINES': str(given_rows),
        'TERM': 'vt220'
    }
    for key, val in mapping.items():
        assert writer.get_extra_info(key) == val


@pytest.mark.anyio
async def test_telnet_server_reject_environ(bind_host,
                                      unused_tcp_port):
    """Test Client's callback method send_environ() for specific requests."""
    from asynctelnet.telopt import SB, NEW_ENVIRON
    # given
    given_cols = 19
    given_rows = 84
    given_encoding = 'cp437'
    given_term = 'vt220'

    class ServerTestEnviron(asynctelnet.TelnetServer):
        def on_request_environ(self):
            return None

    await asynctelnet.create_server(
        protocol_factory=ServerTestEnviron,
        host=bind_host, port=unused_tcp_port)

    reader, writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        cols=given_cols, rows=given_rows, encoding=given_encoding,
        term=given_term, connect_minwait=0.05, connect_maxwait=0.05)

    # this causes the client to expect the server to have demanded environment
    # values, since it did, of course demand DO NEW_ENVIRON! However, our API
    # choice here has chosen not to -- the client then indicates this as a
    # failed sub-negotiation (SB + NEW_ENVIRON).
    _failed = {key: val for key, val in writer.pending_option.items() if val}
    assert _failed == {SB + NEW_ENVIRON: True}



#!/usr/bin/env python3
"""
Telnet Client API for the 'asynctelnet' python package.
"""
# std imports
import argparse
import logging
import codecs
import struct
import anyio
import sys
import os
from contextlib import asynccontextmanager
from functools import partial

# local imports
from . import accessories
from . import client_base

__all__ = ('TelnetClient', 'TelnetTerminalClient', 'open_connection')


class TelnetClient(client_base.BaseClient):
    """
    Telnet client that supports all common options.

    This class is useful for automation, it appears to be a virtual terminal to
    the remote end, but does not require an interactive terminal to run.
    """

    #: On :meth:`send_env`, the value of 'LANG' will be 'C' for binary
    #: transmission.  When encoding is specified (utf8 by default), the LANG
    #: variable must also contain a locale, this value is used, providing a
    #: full default LANG value of 'C.utf8'
    DEFAULT_LOCALE = 'C'

    def __init__(self, conn, term='unknown', cols=80, rows=25,
                 tspeed=(38400, 38400), xdisploc='',
                 *args, **kwargs):
        super().__init__(conn, *args, **kwargs)
        self.__extra = {
            'charset': kwargs['encoding'] or '',
            # for our purposes, we only send the second part (encoding) of our
            # 'lang' variable, CHARSET negotiation does not provide locale
            # negotiation; this is better left to the real LANG variable
            # negotiated as-is by send_env().
            #
            # So which locale should we represent? Rather than using the
            # locale.getpreferredencoding() method, we provide a deterministic
            # class value DEFAULT_LOCALE (en_US), derive and modify as needed.
            'lang': ('C' if not kwargs['encoding'] else
                     self.DEFAULT_LOCALE + '.' + kwargs['encoding']),
            'cols': cols,
            'rows': rows,
            'term': term,
            'tspeed': '{},{}'.format(*tspeed),
            'xdisploc': xdisploc,
        }

    @property
    def extra_attributes(self):
        res = super().extra_attributes
        for k in self.__extra.keys():
            def fn(k):
                return lambda: self.__extra[k] 
            res[k] = fn(k)
        return res

    async def setup(self):
        """Called after setting up."""
        from asynctelnet.telopt import TTYPE, TSPEED, XDISPLOC, NEW_ENVIRON
        from asynctelnet.telopt import CHARSET, NAWS
        await super().setup()

        # Wire extended rfc callbacks for requests of
        # terminal attributes, environment values, etc.
        for (opt, func) in (
                (TTYPE, self.send_ttype),
                (TSPEED, self.send_tspeed),
                (XDISPLOC, self.send_xdisploc),
                (NEW_ENVIRON, self.send_env),
                (NAWS, self.send_naws),
                (CHARSET, self.send_charset),
                ):
            self.set_ext_send_callback(opt, func)

    async def send_ttype(self):
        """Callback for responding to TTYPE requests."""
        return self.__extra['term']

    async def send_tspeed(self):
        """Callback for responding to TSPEED requests."""
        return tuple(map(int, self.__extra['tspeed'].split(',')))

    async def send_xdisploc(self):
        """Callback for responding to XDISPLOC requests."""
        return self.__extra['xdisploc']

    async def send_env(self, keys):
        """
        Callback for responding to NEW_ENVIRON requests.

        :param dict keys: Values are requested for the keys specified.
           When empty, all environment values that wish to be volunteered
           should be returned.
        :returns: dictionary of environment values requested, or an
            empty string for keys not available. A return value must be
            given for each key requested.
        :rtype: dict
        """
        env = {
            'LANG': self._extra['lang'],
            'TERM': self._extra['term'],
            'DISPLAY': self._extra['xdisploc'],
            'LINES': self._extra['rows'],
            'COLUMNS': self._extra['cols'],
        }
        return {key: env.get(key, '') for key in keys} or env

    async def send_charset(self, offered):
        """
        Callback for responding to CHARSET requests.

        Receives a list of character encodings offered by the server
        as ``offered`` such as ``('LATIN-1', 'UTF-8')``, for which the
        client may return a value agreed to use, or None to disagree to
        any available offers.  Server offerings may be encodings or
        codepages.

        The default implementation selects any matching encoding that
        python is capable of using, preferring any that matches
        :py:attr:`encoding` if matched in the offered list.

        :param list offered: list of CHARSET options offered by server.
        :returns: character encoding agreed to be used.
        :rtype: Union[str, None]
        """
        selected = ''
        for offer in offered:
            try:
                codec = codecs.lookup(offer)
            except LookupError as err:
                self.log.info('LookupError: {}'.format(err))
            else:
                if (codec.name == self.default_encoding or not selected):
                    self._extra['charset'] = codec.name
                    self._extra['lang'] = (
                        self.DEFAULT_LOCALE + '.' + codec.name)
                    selected = offer
        if selected:
            self.log.debug('encoding negotiated: {0}'.format(selected))
        else:
            self.log.warning('No suitable encoding offered by server: {!r}.'
                             .format(offered))
        return selected

    async def send_naws(self):
        """
        Callback for responding to NAWS requests.

        :rtype: (int, int)
        :returns: client window size as (rows, columns).
        """
        return (self._extra['rows'], self._extra['cols'])

    def encoding(self, outgoing=None, incoming=None):
        """
        Return encoding for the given stream direction.

        :param bool outgoing: Whether the return value is suitable for
            encoding bytes for transmission to server.
        :param bool incoming: Whether the return value is suitable for
            decoding bytes received by the client.
        :raises TypeError: when a direction argument, either ``outgoing``
            or ``incoming``, was not set ``True``.
        :returns: ``'US-ASCII'`` for the directions indicated, unless
            ``BINARY`` :rfc:`856` has been negotiated for the direction
            indicated or :attr`force_binary` is set ``True``.
        :rtype: str
        """
        if not (outgoing or incoming):
            raise TypeError("encoding arguments 'outgoing' and 'incoming' "
                            "are required: toggle at least one.")

        # may we encode in the direction indicated?
        _outgoing_only = outgoing and not incoming
        _incoming_only = not outgoing and incoming
        _bidirectional = outgoing and incoming
        may_encode = ((_outgoing_only and self.writer.outbinary) or
                      (_incoming_only and self.writer.inbinary) or
                      (_bidirectional and
                       self.writer.outbinary and self.writer.inbinary))

        if self.force_binary or may_encode:
            # The 'charset' value, initialized using keyword argument
            # default_encoding, may be re-negotiated later.  Only the CHARSET
            # negotiation method allows the server to select an encoding, so
            # this value is reflected here by a single return statement.
            return self._extra['charset']
        return 'US-ASCII'


class TelnetTerminalClient(TelnetClient):
    """Telnet client for sessions with a network virtual terminal (NVT)."""

    def send_naws(self):
        """
        Callback replies to request for window size, NAWS :rfc:`1073`.

        :rtype: (int, int)
        :returns: window dimensions by lines and columns
        """
        return self._winsize()

    def send_env(self, keys):
        """
        Callback replies to request for env values, NEW_ENVIRON :rfc:`1572`.

        :rtype: dict
        :returns: super class value updated with window LINES and COLUMNS.
        """
        env = super().send_env(keys)
        env['LINES'], env['COLUMNS'] = self._winsize()
        return env

    @staticmethod
    def _winsize():
        try:
            import fcntl
            import termios
            fmt = 'hhhh'
            buf = '\x00' * struct.calcsize(fmt)
            val = fcntl.ioctl(sys.stdin.fileno(), termios.TIOCGWINSZ, buf)
            rows, cols, _, _ = struct.unpack(fmt, val)
            return rows, cols
        except (ImportError, IOError):
            # TODO: mock import error, or test on windows or other non-posix.
            return (int(os.environ.get('LINES', 25)),
                    int(os.environ.get('COLUMNS', 80)))


@asynccontextmanager
async def open_connection(host=None, port=23, *, log=None, client_factory=None,
                         encoding='utf8', encoding_errors='replace',
                         force_binary=False, term='unknown', cols=80, rows=25,
                         tspeed=(38400, 38400), xdisploc='',
                         connect_minwait=2.0, connect_maxwait=3.0,
                         waiter_closed=None, waiter_connected=None,
                         **kwargs):
    """
    Connect to a TCP Telnet server as a Telnet client.

    :param str host: Remote Internet TCP Server host.
    :param int port: Remote Internet host TCP port.
    :param logging.Logger log: target logger, if None is given, one is created
        using the namespace ``'asynctelnet.server'``.
    :param client_base.BaseClient client_factory: Client connection class
        factory.  When ``None``, :class:`TelnetTerminalClient` is used when
        *stdin* is attached to a terminal, :class:`TelnetClient` otherwise.
    :param str encoding: The default assumed encoding, or ``False`` to disable
        unicode support.  This value is used for decoding bytes received by and
        encoding bytes transmitted to the Server.  These values are preferred
        in response to NEW_ENVIRON :rfc:`1572` as environment value ``LANG``,
        and by CHARSET :rfc:`2066` negotiation.

        The server's attached ``reader, writer`` streams accept and return
        unicode, unless this value explicitly set ``False``.  In that case, the
        attached streams interfaces are bytes-only.

    :param str term: Terminal type sent for requests of TTYPE, :rfc:`930` or as
        Environment value TERM by NEW_ENVIRON negotiation, :rfc:`1672`.
    :param int cols: Client window dimension sent as Environment value COLUMNS
        by NEW_ENVIRON negotiation, :rfc:`1672` or NAWS :rfc:`1073`.
    :param int rows: Client window dimension sent as Environment value LINES by
        NEW_ENVIRON negotiation, :rfc:`1672` or NAWS :rfc:`1073`.
    :param tuple tspeed: Tuple of client BPS line speed in form ``(rx, tx``)
        for receive and transmit, respectively.  Sent when requested by TSPEED,
        :rfc:`1079`.
    :param str xdisploc: String transmitted in response for request of
        XDISPLOC, :rfc:`1086` by server (X11).
    :param Callable shell: A coroutine that is called after
        negotiation completes, receiving arguments ``(reader, writer)``.
        The reader is a :class:`~.TelnetReader` instance, the writer is
        a :class:`~.TelnetWriter` instance.
    :param float connect_minwait: The client allows any additional telnet
        negotiations to be demanded by the server within this period of time
        before launching the shell.  Servers should assert desired negotiation
        on-connect and in response to 1 or 2 round trips.

        A server that does not make any telnet demands, such as a TCP server
        that is not a telnet server will delay the execution of ``shell`` for
        exactly this amount of time.
    :param bool force_binary: When ``True``, the encoding specified is used for
        both directions even when failing ``BINARY`` negotiation, :rfc:`856`.
        This parameter has no effect when ``encoding=False``.
    :param str encoding: Character encoding to use. The default is ``utf-8``.
    :param str encoding_errors: Same meaning as :meth:`codecs.Codec.encode`.
    :param float connect_maxwait: If the remote end is not compliant, or
        otherwise confused by our demands, the shell continues anyway after the
        greater of this value has elapsed.  A client that is not answering
        option negotiation will delay the start of the shell by this amount.

    :return mgr: The reader is a :class:`~.TelnetReader`
        instance, the writer is a :class:`~.TelnetWriter` instance.

    """
    log = log or logging.getLogger(__name__)

    if client_factory is None:
        client_factory = TelnetClient
    async with await anyio.connect_tcp(host, port, **kwargs) as conn:
        async with client_factory(conn, term=term, cols=cols, rows=rows,
                tspeed=tspeed, xdisploc=xdisploc,
                log=log, encoding=encoding, encoding_errors=encoding_errors,
                force_binary=force_binary) as stream:
            yield stream

async def run_client(host, port=23, *, shell=None, **kw):
    async with open_connection(host=host, port=port, **kw) as conn:
        await shell(conn)

def main():
    """Command-line 'asynctelnet-client' entry point, via setuptools."""
    kwargs = _transform_args(_get_argument_parser().parse_args())
    config_msg = (
        'Client configuration: {key_values}'
        .format(key_values=accessories.repr_mapping(kwargs)))
    host = kwargs.pop('host')
    port = kwargs.pop('port')

    log = kwargs['log'] = accessories.make_logger(
        name=__name__,
        loglevel=kwargs.pop('loglevel'),
        logfile=kwargs.pop('logfile'),
        logfmt=kwargs.pop('logfmt'))
    log.debug(config_msg)

    # connect
    anyio.run(partial(run_client,host, port, **kwargs))


def _get_argument_parser():
    parser = argparse.ArgumentParser(
        description="Telnet protocol client",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('host', action='store',
                        help='hostname')
    parser.add_argument('port', nargs='?', default=23, type=int,
                        help='port number')
    parser.add_argument('--term', default=os.environ.get('TERM', 'unknown'),
                        help='terminal type')
    parser.add_argument('--loglevel', default='warn',
                        help='log level')
    parser.add_argument('--logfmt', default=accessories._DEFAULT_LOGFMT,
                        help='log format')
    parser.add_argument('--logfile',
                        help='filepath')
    parser.add_argument('--shell', default='asynctelnet.telnet_client_shell',
                        help='module.function_name')
    parser.add_argument('--encoding', default='utf8',
                        help='encoding name')
    parser.add_argument('--speed', default=38400, type=int,
                        help='connection speed')
    parser.add_argument('--encoding-errors', default='replace',
                        help='handler for encoding errors',
                        choices=('replace', 'ignore', 'strict'))

    parser.add_argument('--force-binary', action='store_true',
                        help='force encoding', default=True)
    parser.add_argument('--connect-minwait', default=1.0, type=float,
                        help='shell delay for negotiation')
    parser.add_argument('--connect-maxwait', default=4.0, type=float,
                        help='timeout for pending negotiation')
    return parser


def _transform_args(args):
    # TODO: Connect as exit(main(**parse_args(sys.argv)))
    return {
        'host': args.host,
        'port': args.port,
        'loglevel': args.loglevel,
        'logfile': args.logfile,
        'logfmt': args.logfmt,
        'encoding': args.encoding,
        'tspeed': (args.speed, args.speed),
        'shell': accessories.function_lookup(args.shell),
        'term': args.term,
        'force_binary': args.force_binary,
        'encoding_errors': args.encoding_errors,
        'connect_minwait': args.connect_minwait,
    }


if __name__ == '__main__':
    exit(main())



#!/usr/bin/python3

import anyio, asynctelnet

async def shell(tcp):
    async with asynctelnet.TelnetServer(tcp) as stream:
        # this will fail if no charset has been negotiated
        await stream.send('\r\nWould you like to play a game? ')
        inp = await reader.receive(1)
        if inp:
            await stream.echo(inp)
            await stream.send('\r\nThey say the only way to win '
                                'is to not play at all.\r\n')

async def main():
    listener = await anyio.create_tcp_listener(local_port=56023)
    await listener.serve(shell)
anyio.run(main)



#!/usr/bin/env python
"""Setuptools distribution file."""
import os
import platform
from setuptools import setup


def _get_here(fname):
    return os.path.join(os.path.dirname(__file__), fname)


def _get_long_description(fname, encoding='utf8'):
    return open(fname, 'r', encoding=encoding).read()


setup(name='asynctelnet',
      # keep in sync w/docs/conf.py manually for now, please!
      use_scm_version={
          "version_scheme": "guess-next-dev",
          "local_scheme": "dirty-tag"
      },
      setup_requires=[
          "setuptools_scm",
      ],
      url='http://asynctelnet.readthedocs.io/',
      license='ISC',
      author='Matthias Urlichs',
      description="Python 3 anyio Telnet server and client Protocol library",
      long_description=_get_long_description(fname=_get_here('README.rst')),
      packages=['asynctelnet'],
      package_data={'': ['README.rst', 'requirements.txt'], },
      entry_points={
         'console_scripts': [
             'asynctelnet-server = asynctelnet.server:main',
             'asynctelnet-client = asynctelnet.client:main'
         ]},
      author_email='matthias@urlichs.de',
      platforms='any',
      zip_safe=True,
      install_requires=[
         'outcome>=1.1',
         'anyio>=2.0.2,<3',
      ],
      keywords=', '.join(('telnet', 'server', 'client', 'bbs', 'mud', 'utf8',
                          'cp437', 'api', 'library', 'anyio', 'trio', 'asyncio', 'talker')),
      classifiers=['License :: OSI Approved :: ISC License (ISCL)',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Intended Audience :: Developers',
                   'Development Status :: 3 - Alpha',
                   'Topic :: System :: Networking',
                   'Topic :: Terminals :: Telnet',
                   'Topic :: System :: Shells',
                   'Topic :: Internet',
                   ],
      )



"""Module provides class BaseServer."""

import traceback
import anyio
import logging
import datetime
import weakref
import sys

from .stream import TelnetStream
from anyio.streams.text import TextStream

__all__ = ('BaseServer',)


class BaseServer(TelnetStream):
    """Base Telnet Server Protocol."""
    _stream_factory = TelnetStream
    _stream_factory_wrapper = TextStream

    def __init__(self, stream, **kw):
        """Class initializer."""
        super().__init__(stream, server=True, **kw)



#!/usr/bin/env python3
"""
A very simple linemode server shell.
"""
# std
import asyncio
import sys
import pkg_resources

# local
import asynctelnet

async def shell(reader, writer):
    from asynctelnet import WONT, ECHO
    writer.iac(WONT, ECHO)

    while True:
        writer.write('> ')

        recv = await reader.readline()

        # eof
        if not recv:
            return

        writer.write('\r\n')

        if recv.rstrip() == 'bye':
            writer.write('goodbye.\r\n')
            await writer.drain()
            writer.close()

        writer.write(''.join(reversed(recv)) + '\r\n')

if __name__ == '__main__':
    kwargs = asynctelnet.parse_server_args()
    kwargs['shell'] = shell
    asynctelnet.run_server(**kwargs)
    #sys.argv.append('--shell={
    sys.exit(
        pkg_resources.load_entry_point(
            'asynctelnet', 'console_scripts', 'asynctelnet-server')()
    )



# Telnet options.
#
# No we don't import from telnetlib. Enums are way nicer.

from enum import IntEnum

__all__ = (
    'Cmd', 'Opt', 'SubT', 'SubVar', 'Flow', 'Req',
    'ABORT', 'ACCEPTED', 'AO', 'AUTHENTICATION', 'AYT', 'BINARY', 'BM',
    'BRK', 'CHARSET', 'CMD_EOR', 'COM_PORT_OPTION', 'DET', 'DM', 'DO',
    'DONT', 'EC', 'ECHO', 'EL', 'ENCRYPT', 'EOF', 'EOR', 'ESC', 'EXOPL',
    'FORWARD_X', 'GA', 'IAC', 'INFO', 'IP', 'IS', 'KERMIT', 'LFLOW',
    'LFLOW_OFF', 'LFLOW_ON', 'LFLOW_RESTART_ANY', 'LFLOW_RESTART_XON',
    'LINEMODE', 'LOGOUT', 'MCCP2_COMPRESS', 'MCCP_COMPRESS', 'NAMS',
    'NAOCRD', 'NAOFFD', 'NAOHTD', 'NAOHTS', 'NAOL', 'NAOLFD', 'NAOP',
    'NAOVTD', 'NAOVTS', 'NAWS', 'NEW_ENVIRON', 'NOP', 'PRAGMA_HEARTBEAT',
    'PRAGMA_LOGON', 'RCP', 'RCTE', 'REJECTED', 'REQUEST', 'RSP', 'SB',
    'SE', 'SEND', 'SEND_URL', 'SGA', 'SNDLOC', 'SSPI_LOGON', 'STATUS',
    'SUPDUP', 'SUPDUPOUTPUT', 'SUPPRESS_LOCAL_ECHO', 'SUSP', 'TLS', 'TM',
    'TN3270E', 'TSPEED', 'TTABLE_ACK', 'TTABLE_IS', 'TTABLE_NAK',
    'TTABLE_REJECTED', 'TTYLOC', 'TTYPE', 'USERVAR', 'VALUE', 'VAR',
    'VT3270REGIME', 'WILL', 'WONT', 'X3PAD', 'XASCII', 'XAUTH',
    'XDISPLOC', 'name_command', 'name_commands',
    'MSSP', 'MSDP', 'GMCP',
)

def _exp(cls):
    for k in dir(cls):
        if k[0].isupper():
            globals()[k] = getattr(cls, k)
    return cls

@_exp
class Cmd(IntEnum):
    EOF = 236  # End of File
    SUSP = 237  # Suspend
    ABORT = 238  # 238
    EOR = 239  # End of Record
    SE = 240  # Subnegotiation Start
    NOP = 241  # No Operation
    DM = 242  # Data Mark
    BRK = 243  # Break
    IP = 244  # Interrupt Process
    AO = 245  # Abort Output
    AYT = 246  # Are You There
    EC = 247  # Erase Character
    EL = 248  # Erase Line
    GA = 249  # Go Ahead
    SB = 250  # Subnegotiation Begin
    WILL = 251  # I want to do …
    WONT = 252  # I will not do …
    DO = 253  # Please do …
    DONT = 254  # You should not do …
    IAC = 255  # Escape

@_exp
class Opt(IntEnum):
    BINARY = 0  # 8-bit data path
    ECHO = 1  # echo
    RCP = 2  # prepare to reconnect
    SGA = 3  # suppress go ahead
    NAMS = 4  # approximate message size
    STATUS = 5  # give status
    TM = 6  # timing mark
    RCTE = 7  # remote controlled transmission and echo
    NAOL = 8  # negotiate about output line width
    NAOP = 9  # negotiate about output page size
    NAOCRD = 10  # negotiate about CR disposition
    NAOHTS = 11  # negotiate about horizontal tabstops
    NAOHTD = 12  # negotiate about horizontal tab disposition
    NAOFFD = 13  # negotiate about formfeed disposition
    NAOVTS = 14  # negotiate about vertical tab stops
    NAOVTD = 15  # negotiate about vertical tab disposition
    NAOLFD = 16  # negotiate about output LF disposition
    XASCII = 17  # extended ascii character set
    LOGOUT = 18  # force logout
    BM = 19  # byte macro
    DET = 20  # data entry terminal
    SUPDUP = 21  # supdup protocol
    SUPDUPOUTPUT = 22  # supdup output
    SNDLOC = 23  # send location
    TTYPE = 24  # terminal type
    EOR = 25  # end or record
    TUID = 26  # TACACS user identification
    OUTMRK = 27  # output marking
    TTYLOC = 28  # terminal location number
    VT3270REGIME = 29  # 3270 regime
    X3PAD = 30  # X.3 PAD
    NAWS = 31  # window size
    TSPEED = 32  # terminal speed
    LFLOW = 33  # remote flow control
    LINEMODE = 34  # Linemode option
    XDISPLOC = 35  # X Display Location
    OLD_ENVIRON = 36  # Old - Environment variables
    AUTHENTICATION = 37  # Authenticate
    ENCRYPT = 38  # Encryption option
    NEW_ENVIRON = 39  # New - Environment variables
    # http://www.iana.org/assignments/telnet-options
    TN3270E = 40  # TN3270E
    XAUTH = 41  # XAUTH
    CHARSET = 42  # CHARSET
    RSP = 43  # Telnet Remote Serial Port
    COM_PORT_OPTION = 44  # Com Port Control Option
    SUPPRESS_LOCAL_ECHO = 45  # Telnet Suppress Local Echo
    TLS = 46  # Telnet Start TLS
    KERMIT = 47  # KERMIT
    SEND_URL = 48  # SEND-URL
    FORWARD_X = 49  # FORWARD_X
    MSDP = 69  # Mud Server Data, https://tintin.mudhalla.net/protocols/msdp/
    MSSP = 70  # Mud Server Status, https://tintin.mudhalla.net/protocols/mssp/
    MCCP_COMPRESS = 85  # MUD compression, old (broken start sequence, DO NOT USE)
    MCCP2_COMPRESS = 86  # MUD compression, new (correct start sequence)
    PRAGMA_LOGON = 138  # TELOPT PRAGMA LOGON
    SSPI_LOGON = 139  # TELOPT SSPI LOGON
    PRAGMA_HEARTBEAT = 140  # TELOPT PRAGMA HEARTBEAT
    GMCP = 201  # Generic Mud Communication Protocol, https://tintin.mudhalla.net/protocols/gmcp/
    EXOPL = 255  # extended options. Not implemented!

CMD_EOR=Cmd.EOR
EOR = Opt.EOR

@_exp
class SubT(IntEnum):
    IS = 0
    SEND = 1
    INFO = 2

@_exp
class SubVar(IntEnum):
    VAR = 0
    VALUE = 1
    ESC = 2
    USERVAR = 3

@_exp
class Flow(IntEnum):
    LFLOW_OFF = 0
    LFLOW_ON = 1
    LFLOW_RESTART_ANY = 2
    LFLOW_RESTART_XON = 3

@_exp
class Req(IntEnum):
    REQUEST = 1
    ACCEPTED = 2
    REJECTED = 3
    TTABLE_IS = 4
    TTABLE_REJECTED = 5
    TTABLE_ACK = 6
    TTABLE_NAK = 7


def name_command(byte):
    """Return string description for (maybe) telnet command byte."""
    try:
        return Cmd(byte)
    except ValueError:
        try:
            return Opt(byte).name
        except ValueError:
            raise ValueError("Unknown: %r"%(byte,)) from None

def name_commands(cmds, sep=' '):
    """Return string description for array of (maybe) telnet command bytes."""
    return sep.join(name_command(byte) for byte in cmds)



# std imports
import asyncio

# 3rd party
import pytest

# local
import asynctelnet
from asynctelnet.tests.accessories import (
    unused_tcp_port,
    bind_host
)


def test_writer_instantiation_safety():
    """On instantiation, one of server or client must be specified."""
    asynctelnet.TelnetWriter(transport=None, protocol=None, client=True)
    with pytest.raises(TypeError):
        # must define at least server=True or client=True
        asynctelnet.TelnetWriter(transport=None, protocol=None)
    with pytest.raises(TypeError):
        # but cannot define both!
        asynctelnet.TelnetWriter(transport=None, protocol=None,
                                server=True, client=True)


def test_repr():
    """Test writer.__repr__ for client and server viewpoint."""
    srv = asynctelnet.TelnetWriter(transport=None, protocol=None, server=True)
    clt = asynctelnet.TelnetWriter(transport=None, protocol=None, client=True)
    assert repr(srv) == ('<TelnetWriter server '
                         'mode:local +lineflow -xon_any +slc_sim>')
    assert repr(clt) == ('<TelnetWriter client '
                         'mode:local +lineflow -xon_any +slc_sim>')


def test_illegal_2byte_iac():
    """Given an illegal 2byte IAC command, raise ValueError."""
    writer = asynctelnet.TelnetWriter(transport=None,
                                     protocol=None,
                                     server=True)
    from asynctelnet.telopt import IAC, SGA
    writer.feed_byte(IAC)
    with pytest.raises(ValueError):
        # IAC SGA(b'\x03'): not a legal 2-byte cmd
        writer.feed_byte(SGA)


def test_legal_2byte_iac():
    """Nothing special about a 2-byte IAC, test wiring a callback."""
    from asynctelnet.telopt import IAC, NOP

    import threading
    called = threading.Event()

    def callback(cmd):
        assert cmd == NOP
        called.set()

    writer = asynctelnet.TelnetWriter(transport=None,
                                     protocol=None,
                                     server=True)

    writer.set_iac_callback(cmd=NOP, func=callback)
    writer.feed_byte(IAC)
    writer.feed_byte(NOP)

    assert called.is_set()


def test_sb_interrupted():
    """IAC SB gets interrupted by IAC command, reseting and exiting state."""
    # when within an SB buffer, all SB protocols we know about remark that
    # IAC must be escaped -- for example, the NAWS negotiation of a 65535
    # by 0 window size should be '\xff\xff\xff\xff\x00\x00' -- so if we
    # receive an IAC **not** followed by an IAC while within a sub-negotiation
    # buffer, we are in miscommunication.  The remote end is not RFC complaint,
    # not a telnet server, or is simply fuzzing us.
    #
    # instead of awaiting the unlikely SE, and throwing all intermediary bytes
    # out, we just clear what we have received so far within this so called
    # 'SB', and exit the sb buffering state.
    from asynctelnet.telopt import IAC, SB, TM, SE
    import logging
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)
    writer = asynctelnet.TelnetWriter(transport=None,
                                     protocol=None,
                                     server=True, log=log)

    given = IAC + SB + b'sbdata-\xff\xff-sbdata'
    sb_expected = b'sbdata-\xff-sbdata'
    for val in given:
        writer.feed_byte(bytes([val]))
    assert b''.join(writer._sb_buffer) == sb_expected

    writer.feed_byte(IAC)
    with pytest.raises(ValueError, match ='SB unhandled'):
        # [SB + b's'] unsolicited,
        writer.feed_byte(SE)

    # the 'IAC TM' interrupts and ends the SB buffer
    given = IAC + SB + b'sbdata-' + IAC + TM + b'-sbdata'
    for val in given:
        writer.feed_byte(bytes([val]))
    assert b''.join(writer._sb_buffer) == b''

    # so, even if you sent an IAC + SE, that is no longer
    # legal for this state.
    writer.feed_byte(b'x')
    writer.feed_byte(IAC)
    with pytest.raises(ValueError, match='not a legal 2-byte cmd'):
        writer.feed_byte(SE)


@pytest.mark.anyio
async def test_iac_do_twice_replies_once(bind_host, unused_tcp_port):
    """WILL/WONT replied only once for repeated DO."""
    from asynctelnet.telopt import IAC, DO, WILL, ECHO

    async def shell(reader, writer):
        writer.close()

    # given,
    given_from_client = IAC + DO + ECHO + IAC + DO + ECHO
    expect_from_server = IAC + WILL + ECHO

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    client_writer.write(given_from_client)

    # verify,
    result_from_server = await asyncio.wait_for(client_reader.read(), 0.5)

    assert result_from_server == expect_from_server


@pytest.mark.anyio
async def test_iac_dont_dont(bind_host, unused_tcp_port):
    """WILL/WONT replied only once for repeated DO."""
    from asynctelnet.telopt import IAC, DONT, ECHO

    async def shell(reader, writer):
        writer.close()

    # given,
    given_from_client = IAC + DONT + ECHO + IAC + DONT + ECHO
    expect_from_server = b''

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    client_writer.write(given_from_client)

    # verify,
    result_from_server = await asyncio.wait_for(client_reader.read(), 0.5)

    assert result_from_server == expect_from_server



@pytest.mark.anyio
async def test_send_iac_dont_dont(bind_host, unused_tcp_port):
    """Try a DONT and ensure it cannot be sent twice."""
    from asynctelnet.telopt import DONT, ECHO
    _waiter = asyncio.Future()

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host,
        port=unused_tcp_port, connect_maxwait=0.05,
        _waiter_connected=_waiter)

    _, client_writer = await asynctelnet.open_connection(
        host=bind_host, port=unused_tcp_port,
        connect_minwait=0.05)

    # say it once,
    result = client_writer.iac(DONT, ECHO)
    assert result == True

    # say it again, suppressed
    result = client_writer.iac(DONT, ECHO)
    assert result == False

    server_writer = (await asyncio.wait_for(_waiter, 0.5)).writer

    assert client_writer.remote_option[ECHO] is False
    assert server_writer.local_option[ECHO] is False


@pytest.mark.anyio
async def test_slc_simul(bind_host, unused_tcp_port):
    """Test SLC control characters are simulated in kludge mode."""
    # For example, ^C is simulated as IP (Interrupt Process) callback.
    from asynctelnet.telopt import IAC, DO, WILL, ECHO, SGA

    # First, change server state into kludge mode -- Then, send all control
    # characters.  We ensure all of our various callbacks that are simulated
    # by control characters were 'fired', as well as the raw bytes received
    # as-is.
    given_input_outband = IAC + DO + ECHO + IAC + DO + SGA
    given_input_inband = bytes(range(ord(' '))) + b'\x7f'
    expected_from_server = IAC + WILL + ECHO + IAC + WILL + SGA
    _waiter_input = asyncio.Future()

    async def shell(reader, writer):
        # read everything from client until they hang up.
        result = await reader.read()

        # then report what was received and hangup on client
        _waiter_input.set_result((writer.protocol.waiters, result))

        writer.close()

    class SimulSLCServer(asynctelnet.BaseServer):
        slc_callbacks = [getattr(asynctelnet.slc, 'SLC_' + key)
                         # no default value for break, sync, or end-of-record.
                         for key in ('IP', 'AO', 'AYT',
                                     'ABORT', 'EOF', 'SUSP', 'EC', 'EL',
                                     'EW', 'RP', 'LNEXT', 'XON', 'XOFF')]

        def connection_made(self, transport):
            super().connection_made(transport)
            self.waiters = {slc_cmd: asyncio.Future()
                            for slc_cmd in self.slc_callbacks}

            for slc_cmd in self.slc_callbacks:
                self.writer.set_slc_callback(
                    slc_byte=slc_cmd,
                    func=lambda byte: self.waiters[byte].set_result(byte))

    await asynctelnet.create_server(
        protocol_factory=SimulSLCServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05,
        encoding=False)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    client_writer.write(given_input_outband)
    client_writer.write(given_input_inband)
    await client_writer.drain()
    result = await client_reader.readexactly(len(expected_from_server))
    assert result == expected_from_server
    client_writer.close()

    # verify,
    callbacks, data_received = await asyncio.wait_for(_waiter_input, 0.5)
    for byte, waiter in callbacks.items():
        assert waiter.done(), asynctelnet.slc.name_slc_command(byte)
    assert data_received == given_input_inband


@pytest.mark.anyio
async def test_unhandled_do_sends_wont(bind_host, unused_tcp_port):
    """An unhandled DO is denied by WONT."""
    from asynctelnet.telopt import IAC, DO, NOP, WONT
    given_input_outband = IAC + DO + NOP
    expected_output = IAC + WONT + NOP

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer,
        host=bind_host, port=unused_tcp_port,
        connect_maxwait=0.05, encoding=False)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # exercise,
    client_writer.write(given_input_outband)
    result = await asyncio.wait_for(
        client_reader.readexactly(len(expected_output)), 0.5)
    assert result == expected_output


@pytest.mark.anyio
async def test_writelines_bytes(bind_host, unused_tcp_port):
    """Exercise bytes-only interface of writer.writelines() function."""
    given = (b'a', b'b', b'c', b'd')
    expected = b'abcd'

    async def shell(reader, writer):
        writer.writelines(given)
        writer.close()

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05,
        encoding=False)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # verify,
    result = await asyncio.wait_for(client_reader.read(), 0.5)

    assert result == expected


@pytest.mark.anyio
async def test_writelines_unicode(bind_host, unused_tcp_port):
    """Exercise unicode interface of writer.writelines() function."""
    given = ('a', 'b', 'c', 'd')
    expected = b'abcd'

    async def shell(reader, writer):
        writer.writelines(given)
        writer.close()

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05,
        encoding='ascii')

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # verify,
    result = await asyncio.wait_for(client_reader.read(), 0.5)

    assert result == expected


def test_bad_iac():
    """Test using writer.iac for something outside of DO/DONT/WILL/WONT."""
    from asynctelnet.telopt import NOP
    writer = asynctelnet.TelnetWriter(transport=None,
                                     protocol=None,
                                     server=True)
    with pytest.raises(ValueError):
        writer.iac(NOP)


@pytest.mark.anyio
async def test_send_ga(bind_host, unused_tcp_port):
    """Writer sends IAC + GA when SGA is not negotiated."""
    from asynctelnet.telopt import IAC, GA
    expected = IAC + GA

    async def shell(reader, writer):
        result = writer.send_ga()
        assert result is True
        writer.close()

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # verify,
    result = await asyncio.wait_for(client_reader.read(), 0.5)

    assert result == expected


@pytest.mark.anyio
async def test_not_send_ga(bind_host, unused_tcp_port):
    """Writer does not send IAC + GA when SGA is negotiated."""
    from asynctelnet.telopt import IAC, DO, WILL, SGA

    # we requires IAC + DO + SGA, and expect a confirming reply.  We also
    # call writer.send_ga() from the shell, whose result should be False
    # (not sent).  The reader never receives an IAC + GA.
    expected = IAC + WILL + SGA

    async def shell(reader, writer):
        result = writer.send_ga()
        assert result is False
        writer.close()

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    client_writer.write(IAC + DO + SGA)

    # verify,
    result = await asyncio.wait_for(client_reader.read(), 0.5)
    assert result == expected


@pytest.mark.anyio
async def test_not_send_eor(bind_host, unused_tcp_port):
    """Writer does not send IAC + EOR when un-negotiated."""
    expected = b''

    async def shell(reader, writer):
        result = writer.send_eor()
        assert result is False
        writer.close()

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # verify,
    result = await asyncio.wait_for(client_reader.read(), 0.5)

    assert result == expected


@pytest.mark.anyio
async def test_send_eor(bind_host, unused_tcp_port):
    """Writer sends IAC + EOR if client requests by DO."""
    from asynctelnet.telopt import IAC, DO, WILL, CMD_EOR, EOR
    given = IAC + DO + EOR
    expected = IAC + WILL + EOR + b'<' + IAC + CMD_EOR + b'>'

    # just verify rfc constants are used appropriately in this context
    assert EOR == bytes([25])
    assert CMD_EOR == bytes([239])

    async def shell(reader, writer):
        writer.write('<')
        result = writer.send_eor()
        assert result is True
        writer.write('>')
        writer.close()

    await asynctelnet.create_server(
        protocol_factory=asynctelnet.BaseServer, host=bind_host, shell=shell,
        port=unused_tcp_port, connect_maxwait=0.05)

    client_reader, client_writer = await asyncio.open_connection(
        host=bind_host, port=unused_tcp_port)

    # verify,
    client_writer.write(given)
    result = await asyncio.wait_for(client_reader.read(), 0.5)

    assert result == expected
